/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.general;

/**
 * A rectangle of maximum 300x300. Its upper left point is included. Its lower
 * right point is not included.
 * @author Niels G. W. Serup
 * @version 1.0
 */
public class Rectangle
{
    private Point upperLeft;
    private Point lowerRight;

    /**
     * Creates a rectangle from two points.
     * @require upperLeft != null
     * @require lowerRight != null
     */
    public Rectangle(Point upperLeft, Point lowerRight)
    {
        if (upperLeft == null || lowerRight == null)
            {
                throw new NullPointerException();
            }
        this.upperLeft = upperLeft;
        this.lowerRight = lowerRight;
    }

    /**
     * Creates a rectangle from two points in number form.
     */
    public Rectangle(int x1, int y1, int x2, int y2)
    {
        this(new Point(x1, y1), new Point(x2, y2));
    }

    /**
     * @return whether or not a given point is inside the rectangle
     */
    public boolean isInside(Point p)
    {
        if (p == null)
            return false;
        return (p.getX() >= this.upperLeft.getX()
                && p.getY() >= this.upperLeft.getY()
                && p.getX() < this.lowerRight.getX()
                && p.getY() < this.lowerRight.getY());
    }

    /**
     * @return whether or not a given rectangle overlaps with this rectangle
     */
    public boolean overlaps(Rectangle other)
    {
        if (other == null)
            return false;
        
        int tSX = this.upperLeft.getX();
        int tEX = this.lowerRight.getX();
        int tSY = this.upperLeft.getY();
        int tEY = this.lowerRight.getY();

        int oSX = other.getUpperLeft().getX();
        int oEX = other.getLowerRight().getX();
        int oSY = other.getUpperLeft().getY();
        int oEY = other.getLowerRight().getY();

        return (oSX < tEX && oEX > tSX && oSY < tEY && oEY > tSY);
    }

    /**
     * Checks if another rectangle touches this rectangle. If this method
     * returns e.g. {@code Direction.NORTH}, it means that the other rectangle
     * is "on top" of this rectangle.
     * @param threshold how many points of "wall" the the rectangles have to
     * share before they are considered to be touching each other. If {@code
     * threshold < 1}, they do not touch.
     * @return the direction at which another rectangle touches this rectangle
     * without overlapping, or {@code null} if no such direction exists.
     */
    public Direction touches(Rectangle other, int threshold)
    {
        if (other == null)
            return null;
        
        int tSX = this.upperLeft.getX();
        int tEX = this.lowerRight.getX();
        int tSY = this.upperLeft.getY();
        int tEY = this.lowerRight.getY();

        int oSX = other.getUpperLeft().getX();
        int oEX = other.getLowerRight().getX();
        int oSY = other.getUpperLeft().getY();
        int oEY = other.getLowerRight().getY();

        if (oSY <= tEY - threshold && oEY >= tSY + threshold)
            {
                if (oSX == tEX)
                    return Direction.EAST;
                if (oEX == tSX)
                    return Direction.WEST;
                return null;
            }
        if (oSX <= tEX - threshold && oEX >= tSX + threshold)
            {
                if (oSY == tEY)
                    return Direction.SOUTH;
                if (oEY == tSY)
                    return Direction.NORTH;
                return null;
            }
        return null;
    }

    /**
     * @return {@code touches(other, 1)}
     */
    public Direction touches(Rectangle other)
    {
        return this.touches(other, 1);
    }

    public Point getUpperLeft()
    {
        return this.upperLeft;
    }

    public Point getLowerRight()
    {
        return this.lowerRight;
    }

    public int getHeight()
    {
        return this.lowerRight.getY() - this.upperLeft.getY();
    }

    public int getWidth()
    {
        return this.lowerRight.getX() - this.upperLeft.getX();
    }

    /**
     * @return width x height
     */
    public int getSize()
    {
        return this.getHeight() * this.getWidth();
    }

    @Override
    public String toString()
    {
        return "(" + this.upperLeft + "," + this.lowerRight + ")";
    }
}
    