/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.gui;
import dungeoncrawl.general.*;
import dungeoncrawl.gui.component.*;
import dungeoncrawl.model.*;
import dungeoncrawl.model.item.*;
import dungeoncrawl.model.player.Player;
import dungeoncrawl.model.player.monster.Monster;
import dungeoncrawl.model.player.hero.*;
import dungeoncrawl.notification.*;
import dungeoncrawl.gui.controller.*;

import java.util.ArrayList;
import java.util.HashMap;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * The dungeon/game panel.
 * 
 * This class completely repaints everything whenever a turn is over. It does
 * this so that the walls of the rooms are not partially overwritten by the
 * edges of circles. To makes things easier, this class overrides JPanel's add*
 * and remove* methods, and the paintComponent method, and implements a simple
 * layer-based paint approach.
 * @see DungeonComponent
 * 
 * @author Niels G. W. Serup
 * @version 1.0
 */
public class DungeonPanel extends JPanel implements Observer<Notification>
{
    public static final Color BACK_COLOR = new Color(0, 0, 0);
    private MainFrame mainFrame;
    private Dungeon dungeon;
    private HashMap<Item, ItemComponent> items;
    private HashMap<Player, PlayerComponent> players;
    private ArrayList<DungeonComponent> components;
    private int dungeonWidth;
    private int dungeonHeight;
    private String[] showingStatus;
    private String finishedText;

    /**
     * Creates a DungeonPanel associated with a MainFrame and a Dungeon.
     */
    public DungeonPanel(MainFrame mainFrame, Dungeon dungeon)
    {
        this.mainFrame = mainFrame;
        this.dungeon = dungeon;
        this.components = new ArrayList<DungeonComponent>();
        this.showingStatus = null;
        this.finishedText = null;
        int[] size = this.dungeon.getSize(true);
        this.dungeonWidth = size[0];
        this.dungeonHeight = size[1];
        dungeon.addObserver(this);
        this.restart();
    }

    public void update(PlayerMoved move)
    {
        // Nothing to do. Wait until TurnIsOver to repaint.
    }

    public void update(TurnIsOver over)
    {
        this.repaint();
        this.mainFrame.requestFocus();
    }

    public void update(PlayerDied death)
    {
        this.remove(this.players.remove(death.player));
    }

    public void update(PickedUpItem pickUp)
    {
        this.remove(this.items.remove(pickUp.item));
    }

    public void update(Finished finished)
    {
        if (finished.heroHasWon)
            this.finishedText = "You won. The monsters lost.";
        else
            this.finishedText = "The monsters won. You lost.";
        this.repaint();
    }

    public boolean isShowingStatus()
    {
        return (this.showingStatus != null);
    }

    public boolean isFinished()
    {
        return (this.finishedText != null);
    }

    public Dungeon getDungeon()
    {
        return this.dungeon;
    }

    public void restart()
    {
        this.dungeon.getHero().placeAtRandomEmptyPosition();
        this.removeAll();
        this.addAll();
    }

    public void add(DungeonComponent component)
    {
        this.components.add(component);
    }

    public void remove(DungeonComponent component)
    {
        this.components.remove(component);
    }

    @Override
    public void removeAll()
    {
        this.components.clear();
    }

    public void addAll()
    {
        // Important to add components in correct order as that will also be
        // the order in which they will be drawn.
        for (Room room : this.dungeon.getRooms())
            {
                this.add(new RoomComponent(room, this));
            }
        for (Room room : this.dungeon.getRooms())
            {
                for (Door door : room.getDoors())
                    {
                        this.add(new DoorComponent(door, this));
                    }
            }

        this.items = new HashMap<Item, ItemComponent>();
        for (Item item : this.dungeon.getItems())
            {
                ItemComponent itemComp = new ItemComponent(item, this);
                this.items.put(item, itemComp);
                this.add(itemComp);
            }

        this.players = new HashMap<Player, PlayerComponent>();
        PlayerComponent playerComp;
        for (Monster monster : this.dungeon.getMonsters())
            {
                playerComp = new PlayerComponent(monster, this);
                this.players.put(monster, playerComp);
                this.add(playerComp);
            }
        Hero hero = this.dungeon.getHero();
        playerComp = new PlayerComponent(hero, this);
        this.players.put(hero, playerComp);
        this.add(playerComp);
    }

    @Override
    public void paintComponent(Graphics g)
        {
            g.setColor(DungeonPanel.BACK_COLOR);
            g.fillRect(0, 0, this.dungeonWidth, this.dungeonHeight);
            for (DungeonComponent component : this.components)
                {
                    component.paintComponent(g);
                }
            if (this.isShowingStatus() || this.isFinished())
                {
                    this.showInfoBox(g);
                }
        }

    public void end()
    {
        this.mainFrame.endGame();
    }

    public void toggleShowingStatus()
    {
        if (this.isShowingStatus())
            {
                this.showingStatus = null;
            }
        else
            {
                this.showingStatus = new String[5];
                Hero hero = this.dungeon.getHero();
                String type = (hero instanceof Warrior ? "warrior"
                               : hero instanceof Mage ? "mage" : "cleric");
                String name = hero.getName() + " (" + type + ")";
                String health = Integer.toString(hero.getHealth());
                String armor = Integer.toString
                    (hero.getArmor().getResistance());
                String damage = Integer.toString(hero.getDamageLevel());
                StringBuilder items = new StringBuilder();
                for (Item item : hero.getInventory())
                    {
                        items.append(" ");
                        items.append(item.toString());
                    }
                this.showingStatus[0] = "Name: " + name;
                this.showingStatus[1] = "Health: " + health;
                this.showingStatus[2] = "Armor: " + armor;
                this.showingStatus[3] = "Damage: " + damage;
                this.showingStatus[4] = "Items:" + items.toString();
            }
        this.repaint();
    }

    /* 
     * Does NOT take into account too small windows, but works for quite small
     * windows in general. If e.g. the hero has picked up too many items, some
     * might be listed outside the window. This could be solved by implementing
     * a scroll functionality.
     */
    private void showInfoBox(Graphics g)
    {
        g.setColor(new Color(0, 255, 100, 190));
        g.fillRect(10, 10, this.dungeonWidth - 20,
                   this.dungeonHeight - 20);
        g.setColor(Color.PINK);
        if (this.finishedText != null)
            {
                g.drawString(this.finishedText, 20, 30);
                g.drawString("Press SPACE to go to", 20, 50);
                g.drawString("the welcome screen.", 20, 65);
            }
        else if (this.showingStatus != null)
            {
                for (int i = 0; i < this.showingStatus.length; i++)
                    {
                        g.drawString(this.showingStatus[i],
                                     20, 30 + 20 * i);
                    }
            }
    }

    public void update(Notification change)
    {
        if (change instanceof PlayerMoved)
            this.update((PlayerMoved) change);
        else if (change instanceof PlayerDied)
            this.update((PlayerDied) change);
        else if (change instanceof PickedUpItem)
            this.update((PickedUpItem) change);
        else if (change instanceof TurnIsOver)
            this.update((TurnIsOver) change);
        else if (change instanceof Finished)
            this.update((Finished) change);
    }
}
