/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.gui;

import java.util.LinkedList;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * The intro panel with a short, hardcoded intro story.
 * @author Niels G. W. Serup
 * @version 1.0
 */
public class IntroPanel extends JPanel
{
    private static final String[][] STORY = new String[][] {
        new String[] {
            "Hello there, {name}. [Press any key to continue.]",
            "[You can skip this intro by pressing ESC.]",
            "No..     No.... NOOOOOOOOOOOOOOOOOO! HELP! HELP! HELP!",
            "<You wake up.>",
            "<There is RPG-style background music.>"
        },
        new String[] {
            "<You are sitting in front of your computer when suddenly a man appears.>",
            "(Or is he not a man...? He looks undeterminably strange.)",
            "Man: {name}! We, the Association of Peace-Keeping Guardians, calleth thou.",
            "Good day. <The man looks away for a moment, as if he was scared.>",
            "Man: Time is short. Come with me.",
            "With you? But I have this programming assignment I have to turn in..",
            "Man: No can do.            <The man claps and both of you disappear.>"
        },
        new String[] {
            "<It is dark. You are thinking about the scream you heard earlier. What was that?>",
            "Man: Can you hear me?",
            "I can't see you, that's for sure.",
            "<The man comes around from a corner with a burning torch.>",
            "Man: How are you, {name}?",
            "I'm fine. But what is this? And the scream? <You touch something slimy.>",
            "Man: You shall see.. You shall see. <He turns around. You follow him.>"
        },
        new String[] {
            "<You have reached a lone computer monitor in the middle of slime and frogs.>",
            "<The computer is running a free GNU/Linux distribution.>",
            "Why is there a computer here?",
            "Man: Oh, we upgraded the dungeon.",
            "DUNGEON??!",
            "Man: Yes, this is our dungeon. Or the monster's dungeon. Our dungeon.",
            "     We have controlled it for centuries, but we are not enough anymore.",
            "     We need YOU to fight the orcs and goblins and free the dungeon!"
        },
        new String[] {
            "But I'm a lazy computer programmer! I can't fight goblins and orcs!",
            "Man: Do you play computer games?",
            "Eh.. Yeah. Duh.",
            "Man: What is your favorite game?",
            "...   Dungeon Crawl of Doom.",
            "Man: So you're the perfect human for the job!"
        },
        new String[] {
            "Man: You see, this computer runs a piece of software that allows you to..",
            "     ..control one of us. We are Heroes, Guardians, Saviours.",
            "     And you shall be our strategist.",
            "But.. If you die...?",
            "Man: We will not die under your control. You are too great. You are {name}!",
            "I know. Yet.. What would happen if, hypothetically, you died?",
            "Man: ... Then you die too. The computer is directly linked to us.",
            "I see."
        },
        new String[] {
            "The description of this dungeon is as follows:",
            "{description}",
            "The red circle is you. Cyan circles are monsters.",
            "Yellow circles are pieces of armor, and magenta circles are weapons.",
            "Green squares are doors.          Man: Are you ready?",
            "   ...     ....        ...     Yes!"
        },
    };
    
    private MainFrame mainFrame;
    private LinkedList<String> strings;
    private String currentString;
    private int currentStringEnd;
    private int frameIndex;
    private int stringIndex;
    private String heroName;
    private String dungeonDescription;
    private Timer letterTimer;

    /**
     * Creates an IntroPanel associated with a MainFrame.
     */
    public IntroPanel(MainFrame mainFrame)
    {
        this.mainFrame = mainFrame;
        this.frameIndex = 0;
        this.stringIndex = 0;
        this.currentStringEnd = 0;
        this.strings = new LinkedList<String>();
        this.heroName = this.mainFrame.getDungeon().getHero().getName();
        this.dungeonDescription = this.mainFrame.getDungeon().getDescription();
        if (this.dungeonDescription == null)
            this.dungeonDescription = "No description.";
        this.letterTimer = new Timer(20, new AbstractAction()
            {
                public void actionPerformed(ActionEvent e)
                {
                    updateCurrentString();
                }
            });
        this.updateStrings();
    }

    public void skip()
    {
        this.mainFrame.endIntro();
    }

    public void updateStrings()
    {
        String string = IntroPanel.STORY[this.frameIndex][this.stringIndex];
        string = string.replace("{name}", this.heroName);
        string = string.replace("{description}", this.dungeonDescription);
        this.currentString = string;
        this.currentStringEnd = 1;
        this.letterTimer.start();
    }

    public void updateCurrentString()
    {
        if (this.currentStringEnd < this.currentString.length())
            {
                this.currentStringEnd++;
                this.repaint();
            }
        else
            {
                this.letterTimer.stop();
            }
    }
    
    public void nextText()
    {
        if (this.letterTimer.isRunning())
            {
                this.letterTimer.stop();
                this.currentStringEnd = this.currentString.length();
                this.repaint();
                return;
            }
        if (this.stringIndex + 1 < IntroPanel.STORY[this.frameIndex].length)
            {
                this.strings.add(this.currentString);
                this.stringIndex++;
                this.updateStrings();
            }
        else
            {
                if (this.frameIndex + 1 < IntroPanel.STORY.length)
                    {
                        this.frameIndex++;
                        this.stringIndex = 0;
                        this.strings.clear();
                        this.updateStrings();
                    }
                else
                    {
                        this.skip();
                        return;
                    }
            }
        this.repaint();
    }

    @Override
    public void paintComponent(Graphics g)
        {
            g.setColor(Color.WHITE);
            g.fillRect(0, 0, this.getWidth(), this.getHeight());
            g.setColor(Color.BLACK);
            int y = 30;
            for (String string : this.strings)
                {
                    g.drawString(string, 20, y);
                    y += 40;
                }
            g.drawString(this.currentString.substring
                         (0, this.currentStringEnd), 20, y);
        }
}
