/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.gui;
import dungeoncrawl.gui.controller.*;
import dungeoncrawl.general.*;
import dungeoncrawl.model.*;
import dungeoncrawl.model.player.hero.*;
import dungeoncrawl.model.player.monster.*;

import java.util.Scanner;
import java.io.File;
import java.io.FileNotFoundException;
import javax.swing.*;
import java.awt.event.KeyListener;
import java.awt.Dimension;

/**
 * The GUI core as well as the main entry point for Dungeon Crawl. I have
 * extended it to allow loading map files from all locations and not just
 * "~/dungeon.map", though it will still load from that file by default.
 * @author Niels G. W. Serup
 * @version 1.0
 */
public class MainFrame extends JFrame
{
    private String mapPath;
    private boolean monstersAreDumb;
    private boolean generateMap;
    private boolean skipIntro;
    private KeyListener keyListener;
    private Dungeon dungeon;
    
    public MainFrame(String mapPath, boolean monstersAreDumb,
                     boolean generateMap, boolean skipIntro)
    {
        super("Dungeon Crawl of Doom");
        this.mapPath = mapPath;
        this.monstersAreDumb = monstersAreDumb;
        this.generateMap = generateMap;
        this.skipIntro = skipIntro;
        this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        this.welcome();
        this.setVisible(true);
    }

    private void welcome()
    {
        this.setSize(640, 400); // Old-school resolution
        this.setResizable(false);
        this.getContentPane().removeAll();
        this.getContentPane().add(new WelcomePanel(this));
        this.validate();
        this.repaint();
        this.requestFocus();
    }

    public KeyListener getKeyListener()
    {
        return this.keyListener;
    }

    public boolean skipsIntro()
    {
        return this.skipIntro;
    }

    public Dungeon getDungeon()
    {
        return this.dungeon;
    }

    private Dungeon generateDungeon()
    {
        return DungeonMapGenerator.generateDungeonMap().getDungeon
            (this.monstersAreDumb);
    }

    private Dungeon readDungeon()
    {
        Dungeon dungeon;
        try
            {
                DungeonMap map;
                if (this.mapPath != null)
                    {
                        map = DungeonParser.parseMap(new Scanner
                                                     (new File(this.mapPath)));
                    }
                else
                    {
                        map = DungeonParser.parseMap();
                    }
                dungeon = map.getDungeon(this.monstersAreDumb);
                dungeon.checkDungeon();
            }
        catch (FileNotFoundException e)
            {
                System.err.println
                    ("dungeoncrawl: error: map file could not"
                     + " be found; generating pseudo-random map..");
                dungeon = this.generateDungeon();
            }
        return dungeon;
    }

    private Dungeon createAlmostEmptyDungeon()
    {
        Dungeon dungeon = new Dungeon(20);
        dungeon.addRoom(new Room(dungeon,
                                 new Rectangle(new Point(0, 0),
                                               new Point(32, 24))));
        dungeon.addMonster(new Goblin(dungeon, new Point(16, 4)));
        return dungeon;
    }

    public void startIntro(String name, String profession)
    {
        if (this.generateMap)
            this.dungeon = this.generateDungeon();
        else
            this.dungeon = this.readDungeon();

        Hero hero;
        if (profession.equals("W"))
            hero = new Warrior(dungeon, name);
        else if (profession.equals("C"))
            hero = new Cleric(dungeon, name);
        else //if (profession.equals("M"))
            hero = new Mage(dungeon, name);
        this.dungeon.setHero(hero);

        if (!this.skipIntro)
            this.switchToIntro();
        else
            this.switchToGame();
    }

    public void switchToIntro()
    {
        final IntroPanel panel = new IntroPanel(this);
        this.keyListener = new IntroKeyListener(panel);
        SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    getContentPane().removeAll();
                    getContentPane().invalidate();
                    getContentPane().add(panel);
                    getContentPane().validate();
                    addKeyListener(keyListener);
                    requestFocus();
                }
            });
    }

    public void switchToGame()
    {
        final DungeonPanel panel = new DungeonPanel(this, dungeon);
        this.keyListener = new DungeonKeyListener(panel);
        SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    getContentPane().removeAll();
                    getContentPane().invalidate();
                    int[] newSize = dungeon.getSize(true);
                    panel.setPreferredSize(new Dimension(newSize[0],
                                                         newSize[1]));
                    getContentPane().add(panel);
                    getContentPane().validate();
                    pack();
                    addKeyListener(keyListener);
                    requestFocus();
                }
            });
    }

    /**
     * End an intro. Assumes an intro has been started.
     */
    public void endIntro()
    {
        SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    removeKeyListener(keyListener);
                    switchToGame();
                }
            });
    }

    /**
     * End a started game. Assumes a game has been started.
     */
    public void endGame()
    {
        SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    removeKeyListener(keyListener);
                    welcome();
                }
            });
    }

    public static void createNewGUI(final String mapPath,
                                    final boolean monstersAreDumb,
                                    final boolean generateMap,
                                    final boolean skipIntro)
    {
        SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    new MainFrame(mapPath, monstersAreDumb, generateMap,
                                  skipIntro);
                }
            });
    }

    /**
     * ALL PROGRAMS SHOULD HAVE A COMMAND-LINE INTERFACE.
     */
    public static void main(String[] args)
    {
        String mapPath = null;
        boolean monstersAreDumb = false;
        boolean generateMap = false;
        boolean skipIntro = false;

        int i = 0;
        while (args.length > i)
            {
                if (args[i].equals("--help") || args[i].equals("-h"))
                    {
                        MainFrame.printHelp();
                        return;
                    }
                if (args[i].equals("--version"))
                    {
                        MainFrame.printVersion();
                        return;
                    }
                if (args[i].equals("-d") || args[i].equals("--dumb"))
                    {
                        monstersAreDumb = true;
                    }
                else if (args[i].equals("-g")
                         || args[i].equals("--generate-map"))
                    {
                        generateMap = true;
                    }
                else if (args[i].equals("-I")
                         || args[i].equals("--skip-intro"))
                    {
                        skipIntro = true;
                    }
                else
                    {
                        mapPath = args[i];
                    }
                i++;
            }
        MainFrame.createNewGUI(mapPath, monstersAreDumb, generateMap,
                               skipIntro);
    }

    private static void printHelp()
    {
        System.out.println
            ("Usage: dungeoncrawl [OPTION]... [MAPFILE]\n"
             + "                   ([MAPFILE] defaults to '~/dungeon.map'.)\n"
             + "\nDungeon Crawl of Doom: The Ultimate Game"
             + " of Boredom\n\n"
             + "Options:\n"
             + "    --version\n"
             + "             show program's version"
             + " number and exit\n"
             + "    -h, --help\n"
             + "             show this help message"
             + " and exit\n"
             + "    -d, --dumb\n"
             + "             monsters do not use the"
             + " improved movement strategy from E.5.1\n"
             + "    -g, --generate-map\n"
             + "             generate a pseudo-random map instead of"
             + " loading a map from a file\n"
             + "    -I, --skip-intro\n"
             + "             do not show the intro");
    }

    private static void printVersion()
    {
        System.out.println("Copyright (C) 2012  Niels G. W. Serup\n"
                           + "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n"
                           + "This is free software: you are free to change and redistribute it.\n"
                           + "There is NO WARRANTY, to the extent permitted by law.");
    }
}
