/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.gui;
import dungeoncrawl.gui.component.*;
import dungeoncrawl.model.*;
import dungeoncrawl.model.player.hero.*;

import java.util.HashMap;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * The welcome panel.
 * @author Niels G. W. Serup
 * @version 1.0
 */
public class WelcomePanel extends JPanel implements ActionListener
{
    public static final Color BACK_COLOR = new Color(255, 221, 0);
    private MainFrame mainFrame;
    private JTextField nameInputField;
    private JRadioButton warriorButton;
    private JRadioButton clericButton;
    private JRadioButton mageButton;
    private ButtonGroup buttonGroup;
    private JLabel regenerationData;
    private JLabel magnitudeData;
    private JButton startButton;

    private boolean fadingOut;
    private int fadedOut;
    private Timer fadeoutTimer;

    private class ProfessionButtonPressed implements ActionListener
    {
        private WelcomePanel welcome;
        private HashMap<String, String> regenData;
        private HashMap<String, String> magData;

        public ProfessionButtonPressed(WelcomePanel welcome)
        {
            this.welcome = welcome;
            this.regenData = new HashMap<String, String>();
            this.magData = new HashMap<String, String>();
            this.regenData.put("W", Integer.toString
                               (Warrior.REGENERATION_LEVEL));
            this.regenData.put("C", Integer.toString
                               (Cleric.REGENERATION_LEVEL));
            this.regenData.put("M", Integer.toString
                               (Mage.REGENERATION_LEVEL));
            this.magData.put("W", Integer.toString
                             (Warrior.DAMAGE_MAGNITUDE));
            this.magData.put("C", Integer.toString
                             (Cleric.DAMAGE_MAGNITUDE));
            this.magData.put("M", Integer.toString
                             (Mage.DAMAGE_MAGNITUDE));
        }

        @Override
        public void actionPerformed(ActionEvent e)
        {   // A radio button was pressed; change the profession info
            String type = ((JRadioButton) e.getSource()).getActionCommand();
            this.welcome.regenerationData.setText(this.regenData.get(type));
            this.welcome.magnitudeData.setText(this.magData.get(type));
        }
    }

    @Override
    // Only the start button triggers this major an event, so it's easier to
    // keep everything in this file.
    public void actionPerformed(ActionEvent e)
    {   // Start button was pressed.
        this.startButton.removeActionListener(this);
        if (!this.mainFrame.skipsIntro())
            this.startFadeout();
        else
            this.gotoIntro();
    }

    public void startFadeout()
    {
        this.fadingOut = true;
        this.fadedOut = 0;
        this.removeAll();

        this.fadeoutTimer = new Timer(10, new AbstractAction()
            {
                public void actionPerformed(ActionEvent e)
                {
                    fadedOut += 10;
                    if (fadedOut > 255)
                        {
                            gotoIntro();
                            return;
                        }
                    repaint();
                }
            });
        this.fadeoutTimer.start();
    }

    public void gotoIntro()
    {
        if (this.fadeoutTimer != null)
            this.fadeoutTimer.stop();
        String name = nameInputField.getText();
        String profession = (buttonGroup.getSelection
                             ().getActionCommand());
        mainFrame.startIntro(name, profession);
    }

    @Override
    public void paintComponent(Graphics g)
    {
        if (this.fadingOut)
            {
                g.setColor(new Color(this.fadedOut, this.fadedOut,
                                     this.fadedOut, this.fadedOut));
                g.fillRect(0, 0, this.getWidth(), this.getHeight());
            }
        else
            super.paintComponent(g);
    }

    public WelcomePanel(MainFrame mainFrame)
    {
        this.mainFrame = mainFrame;

        // GUI-independent structure

        this.nameInputField = new RandomStartValueTextField
            (24, new String[] {"Macbeth", "Banquo", "Niels", "Thor",
                               "The Great", "Babylonian Bubblegum",
                               "Dave", "Blob"});

        ProfessionButtonPressed buttonPressed
            = new ProfessionButtonPressed(this);
        this.warriorButton = new JRadioButton("Warrior");
        this.warriorButton.setActionCommand("W");
        this.warriorButton.addActionListener(buttonPressed);

        this.clericButton = new JRadioButton("Cleric");
        this.clericButton.setActionCommand("C");
        this.clericButton.addActionListener(buttonPressed);
        
        this.mageButton = new JRadioButton("Mage");
        this.mageButton.setActionCommand("M");
        this.mageButton.addActionListener(buttonPressed);
        
        this.regenerationData = new JLabel();
        this.magnitudeData = new JLabel();

        this.clericButton.setSelected(true);
        buttonPressed.actionPerformed
            (new ActionEvent((Object) this.clericButton, 0, ""));

        this.startButton = new JButton("Start");
        this.startButton.addActionListener(this);
        

        // Actual layout

        this.setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
        
        this.setBackground(WelcomePanel.BACK_COLOR);

        JTextPane textPane = new JTextPane();
        textPane.setBackground(WelcomePanel.BACK_COLOR);
        textPane.setContentType("text/html");
        textPane.setEditable(false);
        textPane.setText
            ("<h1>Dungeon Crawl of Doom</h1>"
             + "<p>Lo! Art thou the strongest, all's rebel's bravedom;<br \\>"
             + "by the sources of insanity, shall you conquer?</p>"
             + "<p>As 'twere eyed life in the deepest of bubbles<br \\>"
             + "shall thy spur be fairer than Heaven!</p>");
        JPanel textPanePanel = new JPanel();
        textPanePanel.setBackground(WelcomePanel.BACK_COLOR);
        textPanePanel.add(textPane);
        this.add(textPanePanel);

        JPanel nameInput = new JPanel();
        JLabel label = new JLabel("Thy name:");
        nameInput.add(label);
        nameInput.add(this.nameInputField);
        nameInput.setBackground(WelcomePanel.BACK_COLOR);
        this.add(nameInput);


        // Click on one, unclick the previously clicked
        this.buttonGroup = new ButtonGroup();
        this.buttonGroup.add(this.warriorButton);
        this.buttonGroup.add(this.clericButton);
        this.buttonGroup.add(this.mageButton);

        JPanel radioPanel = new JPanel(new GridLayout(0, 1));
        radioPanel.add(this.warriorButton);
        radioPanel.add(this.clericButton);
        radioPanel.add(this.mageButton);
        this.warriorButton.setBackground(WelcomePanel.BACK_COLOR);
        this.clericButton.setBackground(WelcomePanel.BACK_COLOR);
        this.mageButton.setBackground(WelcomePanel.BACK_COLOR);

        JPanel profPanel = new JPanel();
        profPanel.add(new JLabel("Thy profession:"));
        profPanel.add(Box.createHorizontalStrut(228));
        profPanel.setBackground(WelcomePanel.BACK_COLOR);
        this.add(profPanel);

        JLabel regenerationLabel = new JLabel("Regeneration: ");
        JLabel magnitudeLabel = new JLabel("Damage magnitude: ");
        regenerationLabel.setPreferredSize(new Dimension(180, 20));
        this.regenerationData.setPreferredSize(new Dimension(40, 20));
        magnitudeLabel.setPreferredSize(new Dimension(180, 20));
        this.magnitudeData.setPreferredSize(new Dimension(40, 20));
        
        JPanel regenerationPanel = new JPanel();
        JPanel flow = new JPanel();
        flow.setBackground(WelcomePanel.BACK_COLOR);
        flow.add(regenerationLabel);
        flow.add(this.regenerationData);
        regenerationPanel.setLayout(new BoxLayout(regenerationPanel, BoxLayout.Y_AXIS));
        regenerationPanel.add(flow);

        JPanel magnitudePanel = new JPanel();
        flow = new JPanel();
        flow.setBackground(WelcomePanel.BACK_COLOR);
        flow.add(magnitudeLabel);
        flow.add(this.magnitudeData);
        magnitudePanel.setLayout(new BoxLayout(magnitudePanel, BoxLayout.Y_AXIS));
        magnitudePanel.add(flow);

        regenerationPanel.setBackground(WelcomePanel.BACK_COLOR);
        magnitudePanel.setBackground(WelcomePanel.BACK_COLOR);
        regenerationLabel.setBackground(WelcomePanel.BACK_COLOR);
        this.regenerationData.setBackground(WelcomePanel.BACK_COLOR);
        magnitudeLabel.setBackground(WelcomePanel.BACK_COLOR);
        this.magnitudeData.setBackground(WelcomePanel.BACK_COLOR);
        JPanel infoPanel = new JPanel(new GridLayout(0, 1));
        infoPanel.add(regenerationPanel);
        infoPanel.add(magnitudePanel);
        infoPanel.setBackground(WelcomePanel.BACK_COLOR);
        infoPanel.setAlignmentX(Component.LEFT_ALIGNMENT);

        JPanel choosePanel = new JPanel();
        choosePanel.setBackground(WelcomePanel.BACK_COLOR);
        choosePanel.add(radioPanel);
        choosePanel.add(Box.createHorizontalStrut(20));
        choosePanel.add(infoPanel);

        this.add(choosePanel);

        this.startButton.setBackground(Color.BLACK);
        this.startButton.setForeground(Color.WHITE);
        this.add(this.startButton);
        this.add(Box.createVerticalStrut(20));

        // Swing was not written for humans.
    }
}
