/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.model;
import dungeoncrawl.general.*;
import dungeoncrawl.model.player.Player;
import dungeoncrawl.model.player.monster.Monster;
import dungeoncrawl.model.player.hero.Hero;
import dungeoncrawl.model.item.Item;
import dungeoncrawl.notification.*;

import java.util.Collection;
import java.util.Comparator;
import java.util.Collections;
import java.util.List;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.LinkedList;

/**
 * A dungeon. Elements of a dungeon include: rooms (which have doors), items,
 * and players.
 * 
 * Rooms cannot be changed once they have been created. Items can be removed
 * when a hero picks them up, and players can die.
 * 
 * This class is observable and will spout out notifications when something
 * happens:
 * 
 * # If an item is picked up, a PickedUpItem notification is issued.
 * # If a player dies, a PlayerDied notification is sent out.
 * # When both the hero and the monsters have made their moves, a TurnIsOver
 *   notification is sent out.
 * # If either the hero is dead or all monsters are dead, the game is
 *   considered finished, and a Finished notification is sent out.
 * 
 * A fifth notification, PlayerMoved, is sent out after a player move, but that
 * happens in the Player class.
 * 
 * @author Niels G. W. Serup
 * @version 1.0
 */
public class Dungeon extends Observable<Notification, Observer<Notification>>
{
    public static final int DEFAULT_SCALE_FACTOR = 20;
    private HashMap<Point, Monster> monsters;
    private Hero hero;
    private HashMap<Point, Item> items;
    private ArrayList<Room> rooms;

    private final int scaleFactor;
    private String description;
    
    /**
     * Creates an empty dungeon.
     */
    public Dungeon(int scaleFactor, String description)
    {
        super();
        this.monsters = new HashMap<Point, Monster>();
        this.items = new HashMap<Point, Item>();
        this.rooms = new ArrayList<Room>();
        this.description = description;
        this.scaleFactor = scaleFactor;
        this.hero = null;
    }

    /**
     * Creates a new dungeon with no description.
     */
    public Dungeon(int scaleFactor)
    {
        this(scaleFactor, null);
    }

    /**
     * Creates a new dungeon with a default scale factor and no description.
     */
    public Dungeon()
    {
        this(Dungeon.DEFAULT_SCALE_FACTOR);
    }


    
    public int getScaleFactor()
    {
        return this.scaleFactor;
    }

    public String getDescription()
    {
        return this.description;
    }

    /**
     * @param scaled whether or not the return size should be scaled with the
     * scale factor
     * @return the size of the dungeon
     */
    public int[] getSize(boolean scaled)
    {
        // All dungeons start in (0, 0). All we need to do is find the most far
        // away lower right point.
        int maxX = 0;
        int maxY = 0;
        int tmpX, tmpY;
        for (Room room : this.rooms)
            {
                tmpX = room.getBoundary().getLowerRight().getX();
                tmpY = room.getBoundary().getLowerRight().getY();
                if (tmpX > maxX)
                    maxX = tmpX;
                if (tmpY > maxY)
                    maxY = tmpY;
            }
        if (scaled)
            {
                maxX *= this.scaleFactor;
                maxY *= this.scaleFactor;
            }
        return new int[] { maxX, maxY };
    }

    /**
     * Checks that the dungeon is not above 300x300 points. Either this method
     * or checkSize() should always be called after adding rooms, except if
     * isMalformed() or checkDungeon() is called, or if the dungeon has been
     * created by DungeonMapGenerator.
     * @return {@code true} if the dungeon map is too large; {@code false}
     * otherwise
     */
    public boolean isTooLarge()
    {
        int[] wh = this.getSize(false);
        return (wh[0] > 300 || wh[1] > 300);
    }

    /**
     * Checks that the the dungeon follows the rules laid out in
     * DungeonMapGenerator.
     *
     * It is not imperative that this method is called before a game is
     * started; a dungeon can be playable even if it does not follow the
     * rules. Note that if the dungeon has a large map, the checking can take
     * some time.
     * @assume the hero has not been added to the dungeon (i.e. {@code
     * this.hero == null})
     * @return {@code true} if the dungeon is malformed; {@code false}
     * otherwise
     */
    public boolean isMalformed()
    {
        return (this.isTooLarge() || this.roomsOverlap()
                || this.itemsOverlapOrNotInRoom()
                || this.monstersOverlapOrNotInRoom()
                || this.hasRogueRoom());
    }

    /**
     * Either this method or isTooLarge() should always be called after adding
     * rooms, except if isMalformed() or checkDungeon() is called, or if the
     * dungeon has been created by DungeonMapGenerator.
     * @throws MapTooLarge if the dungeon is too large
     * @see #isTooLarge()
     */
    public void checkSize()
    {
        if (this.isTooLarge())
            throw new MapTooLarge();
    }

    /**
     * @throws MalformedDungeon if the dungeon is malformed
     * @see #isMalformed()
     */
    public void checkDungeon()
    {
        if (this.isMalformed())
            throw new MalformedDungeon();
    }

    /**
     * @return {@code true} if at least one room overlaps with at least one
     * other room; {@code false} otherwise
     */
    public boolean roomsOverlap()
    {
        for (Room room : this.rooms)
            {
                for (Room otherRoom : this.rooms)
                    {
                        if (otherRoom != room && room.getBoundary().overlaps
                            (otherRoom.getBoundary()))
                            return true;
                    }
            }
        return false;
    }

    /**
     * @return {@code true} if a given position is inside the dungeon; {@code
     * false} otherwise
     */
    public boolean isInside(Point position)
    {
        for (Room room : this.rooms)
            {
                if (room.isInside(position))
                    return true;
            }
        return false;
    }

    /**
     * @return {@code true} if either at least two items overlap or at least
     * one item is outside the dungeon; {@code false} otherwise
     */
    public boolean itemsOverlapOrNotInRoom()
    {
        for (Item item : this.getItems())
            {
                if (!this.isInside(item.getPosition()))
                    return true;
                for (Item otherItem : this.getItems())
                    {
                        if (otherItem != item && item.getPosition().equals
                            (otherItem.getPosition()))
                            return true;
                    }
            }
        return false;
    }

    /**
     * @return {@code true} if either at least two monsters overlap or at least
     * one monster is outside the dungeon; {@code false} otherwise
     */
    public boolean monstersOverlapOrNotInRoom()
    {
        for (Monster monster : this.getMonsters())
            {
                if (!this.isInside(monster.getPosition()))
                    return true;
                for (Monster otherMonster : this.getMonsters())
                    {
                        if (otherMonster != monster && monster.getPosition
                            ().equals(otherMonster.getPosition()))
                            return true;
                    }
            }
        return false;
    }

    /**
     * Note: This method could probably be optimized.
     * @return {@code true} if at least one room does not allow for complete
     * passage to every other room; {@code false} otherwise
     */
    public boolean hasRogueRoom()
    {
        if (this.rooms.size() < 1)
            return false;
        if (this.rooms.size() < 2)
            return true;
        Room firstRoom = this.rooms.get(0); // All doors are two-way, so if
                                            // there is a path from the first
                                            // room to every other room, the
                                            // same is true for the other
                                            // rooms.
        for (Room room : this.rooms.subList(1, this.rooms.size()))
            {
                if (!firstRoom.hasPath(room))
                    return true;
            }
        return false;
    }

    

    public ArrayList<Room> getRooms()
    {
        return this.rooms;
    }

    public Hero getHero()
    {
        return this.hero;
    }

    public Collection<Monster> getMonsters()
    {
        return this.monsters.values();
    }

    public Collection<Item> getItems()
    {
        return this.items.values();
    }

    public Item getItemAt(Point position)
    {
        return this.items.get(position);
    }

    public Player getPlayerAt(Point p)
    {
        if (this.hero != null && p.equals(this.hero.getPosition()))
            {
                return this.hero;
            }
        return this.monsters.get(p);
    }

    /**
     * @param p the point to test for
     * @return the room where a point is inside, or {@code null} if no such
     * room exists
     */
    public Room getRoomOfPosition(Point p)
    {
        for (Room room : this.rooms)
            {
                if (room.isInside(p))
                    {
                        return room;
                    }
            }
        return null;
    }

    /**
     * Adds a room to the dungeon. Assumes it is valid and does not overlap.
     */
    public void addRoom(Room room)
    {
        this.rooms.add(room);
    }

    /**
     * Adds two one-way doors to the two rooms where the two adjacent points
     * {@code p1} and {@code p2} exist, {@code p1} in room 1 and {@code p2} in
     * room 2, if such two rooms exist.
     */
    public void addDoor(Point p1, Point p2)
    {
        if (p1 != null && p2 != null)
            {
                Room room1 = this.getRoomOfPosition(p1);
                Room room2 = this.getRoomOfPosition(p2);
                room1.addDoor(p1, p2, room2);
                room2.addDoor(p2, p1, room1);
            }
    }

    /**
     * Adds a monster to the dungeon.
     * @assume monster != null
     */
    public void addMonster(Monster monster)
    {
        this.monsters.put(monster.getPosition(), monster);
    }

    /**
     * Sets the hero of the dungeon.
     * @assume hero != null;
     */
    public void setHero(Hero hero)
    {
        this.hero = hero;
    }

    public void addItem(Item item)
    {
        this.items.put(item.getPosition(), item);
    }

    /**
     * Removes a given player if present.
     */
    public void removePlayer(Player player)
    {
        if (player == this.hero)
            this.hero = null;
        else
            this.monsters.remove(player.getPosition());
    }

    /**
     * Removes a given monster if present.
     */
    public void removeMonster(Monster monster)
    {
        this.monsters.remove(monster.getPosition());
    }

    public Item loot(Point position)
    {
        Item item = this.items.remove(position);
        if (item != null)
            {
                this.notifyObservers(new PickedUpItem(item));
            }
        return item;
    }

    public void updateMonsterPosition(Monster monster, Point newPosition)
    {
        this.removeMonster(monster);
        this.monsters.put(newPosition, monster);
    }

    /**
     * Checks if there is a wall between two adjacent points.
     * @assume p1 != null && p2 != null
     * @return whether or not there is a wall between {@code p1} and {@code p2}
     */
    public boolean isWallBetween(Point p1, Point p2)
    {
        return (!this.isClearPath(p1, p2));
    }

    /**
     * Does the opposite of {@code isWallBetween(Point p1, Point p2)}.
     * @assume p1 != null && p2 != null
     * @return whether or not there are no walls in the way between {@code p1}
     * and {@code p2}
     */
    public boolean isClearPath(Point p1, Point p2)
    {
        Room room1 = this.getRoomOfPosition(p1);
        Room room2 = this.getRoomOfPosition(p2);
        return ((room1 == room2)
                || (room1.getDoorAt(p1) != null
                    && room2.getDoorAt(p2) != null));
    }

    /**
     * This is where everything happens after the hero has moved.
     */
    public void afterHeroMove()
    {
        ArrayList<Monster> monsters = this.moveMonstersAutonomously();
        boolean heroIsDead = this.hero.isDead();
        boolean allMonstersAreDead = true; // We assume.
        if (heroIsDead)
            {
                this.notifyObservers(new PlayerDied(this.hero));
                this.hero = null;
            }
        else
            {
                this.hero.regenerate();
                
                for (Monster monster : monsters)
                    {
                        if (monster.isDead())
                            {
                                this.notifyObservers(new PlayerDied(monster));
                                this.removeMonster(monster);
                            }
                        else
                            {
                                allMonstersAreDead = false;
                                monster.regenerate();
                            }
                    }
            }
        this.notifyObservers(new TurnIsOver());
        if (heroIsDead)
            {
                this.notifyObservers(new Finished(false));
            }
        else if (allMonstersAreDead)
            {
                this.notifyObservers(new Finished(true));
            }
    }

    /*
     * Move all monsters as if they moved together by themselves.
     *
     *
     * The naive thing to do would to mindlessly ask each monster to move. This
     * could have a negative impact on the fluidity of the game. For example,
     * consider the following situation where there are two monsters, 1 and 2,
     * and our hero, H:
     *
     *          H 12
     * 
     * Since the first part of a monster's job is to move closer to the hero,
     * both 1 and 2 will try to move left, like this:
     *
     *          H12
     * 
     * But monsters may not hit each other, so depending on the order in which
     * 1 and 2 are moved, this might also happen:
     *
     *          H1 2
     * 
     * This would be the case if 2 was moved before 1 in a naive monster
     * movement implementation. In the former case, 1 was moved before 2, so 2
     * did not risk hitting 1 and could move left just as 1 could; in this case
     * 2 cannot move left because then it would hit 1, but 1 can move
     * left. This is not desirable.
     * 
     * To avoid this, I see two solutions which both allow for a more intuitive
     * way of moving.
     * 
     * The goal should be for all monster movement to appear swift, as if they
     * were one large organism, independent of the order in which they are
     * stored.
     * 
     **************************************************************************
     * 1) Start with flagging all monsters that they are ready to move, then
     *    move them.
     * 
     * Flag all monsters before they are moved, so that a monster can ask
     * another monster to move (and so on recursively) before it itself
     * moves. If one did not flag monsters, some might move twice in one turn
     * under this approach.
     * 
     * NOTE 1: This flag approach does not better a movement strategy where
     * movements are determined at random.
     * 
     * NOTE 2: If there are many monsters, this approach might cause stack
     * overflows.
     * 
     * NOTE 3: This approach also requires more memory. If for some reason a
     * monster 1 wanted to move towards a monster 2, and 2 wanted to move
     * towards 1, and they were next to each other, there would be a stack
     * overflow.
     **************************************************************************
     * 2) Sort the list of monsters after how close they are to the hero.
     * 
     * The monster who thinks it is closest to the hero moves first.
     * 
     * NOTE 1: This flag approach does not better a movement strategy where
     * movements are determined at random.
     * 
     * NOTE 2: The monster who thinks it is closest might not be closest at
     * all, if that monster uses a poorer movement strategy than another
     * monster who is actually closer but who uses a better movement strategy.
     *
     * This is the approach I have chosen. It is more resistant to random
     * movement strategies and uses less stack and memory.
     */
    // Returns a list of current monsters.
    @SuppressWarnings("unchecked") // Due to the Comparator object being used
    private ArrayList<Monster> moveMonstersAutonomously()
    {
        // Create a new list to avoid ConcurrentModificationException.
        ArrayList<Monster> monsters = new ArrayList<Monster>();
        for (Monster monster : this.monsters.values())
            {
                monsters.add(monster);
            }
        // Sort the new list.
        Collections.sort(monsters, new Comparator() {
                public int compare(Object o1, Object o2)
                {
                    return (((Monster) o1).getHeroApproximity()
                            - ((Monster) o2).getHeroApproximity());
                }
                public boolean equals(Object obj) {return false;}
            });
        
        for (Monster monster : monsters)
            {
                monster.makeAutonomousMove();
            }
        return monsters;
    }
}
