/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.model;
import dungeoncrawl.general.*;
import dungeoncrawl.model.player.monster.*;
import dungeoncrawl.model.item.*;
import dungeoncrawl.model.movementstrategy.*;

import java.util.ArrayList;

/**
 * An internal storage of the data also found in .map files. The purpose of
 * this class is to be able to play the same level over and over again without
 * having to reload it from disk over and over again.
 * @author Niels G. W. Serup
 * @version 1.0
 */
public class DungeonMap
{
    public class PositionAndStrength
    {
        public Point position;
        public int strength;
        public PositionAndStrength(Point position, int strength)
        {
            this.position = position;
            this.strength = strength;
        }
    }

    private String description;
    private int scaleFactor;
    private ArrayList<Rectangle> rooms;
    private ArrayList<Point[]> doors;
    private ArrayList<Point> orcs;
    private ArrayList<Point> goblins;
    private ArrayList<PositionAndStrength> armor;
    private ArrayList<PositionAndStrength> weapons;

    private void initFields()
    {
        this.rooms = new ArrayList<Rectangle>();
        this.doors = new ArrayList<Point[]>();
        this.orcs = new ArrayList<Point>();
        this.goblins = new ArrayList<Point>();
        this.armor = new ArrayList<PositionAndStrength>();
        this.weapons = new ArrayList<PositionAndStrength>();
    }
    
    /**
     * Creates a new map.
     */
    public DungeonMap(int scaleFactor, String description)
    {
        this.description = description;
        this.scaleFactor = scaleFactor;
        this.initFields();
    }

    /**
     * Creates a new map from a dungeon. This can be useful for two purposes:
     * 
     * 1) A dungeon is easier to modify than a dungeon map. For example,
     * DungeonMapGenerator works on a dungeon, but returns a dungeon map.
     * 
     * 2) It makes it possible to save the current game (almost --- the hero
     * and data about whether or not monsters are dumb are not saved, but those
     * features could be added if they were needed).
     * 
     * Does not modify the Dungeon.
     */
    public DungeonMap(Dungeon dungeon)
    {
        this.scaleFactor = dungeon.getScaleFactor();
        this.description = dungeon.getDescription();
        this.initFields();
        Point doorStart, doorEnd;
        for (Room room : dungeon.getRooms())
            {
                this.rooms.add(room.getBoundary());
                for (Door door : room.getDoors())
                    {
                        doorStart = door.getPosition();
                        doorEnd = door.getTargetPosition();
                        if (!this.doorsHaveBeenAdded(doorEnd))
                            this.doors.add(new Point[] { doorStart, doorEnd });
                    }
            }
        for (Monster monster : dungeon.getMonsters())
            {
                if (monster instanceof Goblin)
                    this.goblins.add(monster.getPosition());
                else if (monster instanceof Orc)
                    this.orcs.add(monster.getPosition());
            }
        for (Item item : dungeon.getItems())
            {
                if (item instanceof Armor)
                    this.armor.add(new PositionAndStrength
                                   (item.getPosition(),
                                    ((Armor) item).getResistance()));
                else if (item instanceof Weapon)
                    this.weapons.add(new PositionAndStrength
                                     (item.getPosition(),
                                      ((Weapon) item).getDamage()));
            }
    }

    private boolean doorsHaveBeenAdded(Point doorEnd)
    {
        for (Point[] otherDoors : this.doors)
            {
                if (otherDoors[0] == doorEnd)
                    return true;
            }
        return false;
    }

    public String getDescription()
    {
        return this.description;
    }

    public int getScaleFactor()
    {
        return this.scaleFactor;
    }

    public ArrayList<Rectangle> getRooms()
    {
        return this.rooms;
    }

    public ArrayList<Point[]> getDoors()
    {
        return this.doors;
    }

    public ArrayList<Point> getOrcs()
    {
        return this.orcs;
    }

    public ArrayList<Point> getGoblins()
    {
        return this.goblins;
    }

    public ArrayList<PositionAndStrength> getArmor()
    {
        return this.armor;
    }

    public ArrayList<PositionAndStrength> getWeapons()
    {
        return this.weapons;
    }

    public void addRoom(Point upperLeft, Point lowerRight)
    {
        this.rooms.add(new Rectangle(upperLeft, lowerRight));
    }

    public void addDoor(Point p1, Point p2)
    {
        this.doors.add(new Point[] { p1, p2 });
    }

    public void addOrc(Point position)
    {
        this.orcs.add(position);
    }

    public void addGoblin(Point position)
    {
        this.goblins.add(position);
    }

    public void addArmor(Point position, int strength)
    {
        this.armor.add(new PositionAndStrength(position, strength));
    }

    public void addWeapon(Point position, int strength)
    {
        this.weapons.add(new PositionAndStrength(position, strength));
    }

    /**
     * Creates a new Dungeon from the map where all monsters move
     * stupidly. Modifying the new Dungeon will not modify this map.
     * @return the Dungeon
     */
    public Dungeon getDungeon(boolean monstersAreDumb)
    {
        Dungeon dungeon = new Dungeon(this.scaleFactor, this.description);
        for (Rectangle rect : this.rooms)
            {
                dungeon.addRoom(new Room(dungeon, rect));
            }
        dungeon.checkSize();
        for (Point[] ps : this.doors)
            {
                dungeon.addDoor(ps[0], ps[1]);
            }
        for (Point orcPos : this.orcs)
            {
                dungeon.addMonster
                    (new Orc(dungeon, orcPos,
                             monstersAreDumb ? new ManhattanMovementStrategy()
                             : new BetterMovementStrategy()));
            }
        for (Point goblinPos : this.goblins)
            {
                dungeon.addMonster
                    (new Goblin(dungeon, goblinPos,
                                monstersAreDumb ? new ManhattanMovementStrategy()
                                : new BetterMovementStrategy()));
            }
        for (DungeonMap.PositionAndStrength ps : this.armor)
            {
                dungeon.addItem(new Armor(dungeon, ps.strength, ps.position));
            }
        for (DungeonMap.PositionAndStrength ps : this.weapons)
            {
                dungeon.addItem(new Weapon(dungeon, ps.strength, ps.position));
            }
        return dungeon;
    }

    /**
     * Creates a new Dungeon from the map where all monsters move
     * cleverly. Modifying the new Dungeon will not modify this map.
     * @return the Dungeon
     */
    public Dungeon getDungeon()
    {
        return this.getDungeon(false);
    }
}
