/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.model;
import dungeoncrawl.general.*;
import dungeoncrawl.model.player.monster.*;
import dungeoncrawl.model.item.*;
import dungeoncrawl.model.movementstrategy.*;

import java.util.Random;
import java.util.ArrayList;

/**
 * An pseudo-random dungeon map generator.
 *
 * Room-related rules:
 *
 * 0. A dungeon is at max 300x300 points large.
 * 1. Every room has at least one door.
 * 2. Doors are located on the walls of the room, but never in a corner.
 * 3. There should be a path from every room to every other room.
 * 4. Rooms may not overlap.
 * 5. A room is at minimum 3x3 points large.
 * 
 * 
 * Method for creating a maximum of {@code n} randomly-sized rooms filling
 * ca. {@code m} points in a dungeon connected to each other in an
 * {@code a}x{@code b}-sized dungeon with a error threshold of {@code t}:
 * 
 * 1. Create a room. Place it somewhere in the dungeon.
 * 
 * 2. If {@code n} rooms have been created, or if {@code m} points have been
 * filled with rooms, or if this is the {@code t}th time the creation of a new
 * room has been attempted without succes, go to step 7. Else, go to step 3.
 * 
 * 3. Create a new room.
 *
 * 4. Try to place it so that a door between it and one of the former rooms can
 * be made. If this is not possible, redo step 2, but with a smaller room, if a
 * smaller room can be created, else go to step 7.
 * 
 * 5. Place one door between the new room and the former room. If the new room
 * touches any other former rooms on at least 3 points, maybe add doors there
 * as well.
 * 
 * 6. Repeat step 2.
 * 
 * 7. Add monsters and items. Two monsters may not share the same position. Two
 * items may not share the same position. A monster and a item may share the
 * same position.
 * 
 * 
 * I call this "The Simple Dungeon-Generation Algorithm", or SDGA.
 ******************************************************************************
 * 
 * How to place a new room that touches one or more former rooms:
 * 
 * This is a topic not covered in SDGA. This is because there are several ways
 * to do this. Either:
 * 
 * 1) Try all available positions in some pseudo-random order, and choose the
 * first one where the new room does not overlap with any former rooms, and
 * where it touches at least one room (without overlapping).
 * 
 * 2) Continually try to place the new room "around" a former room. Think of
 * this as first attempting to slide the new room on top of the former room,
 * trying to find a place where it does not overlap with other rooms. If that
 * does not succeed, think of sliding it on the right side of a former room,
 * and so on, until maybe an empty spot is found.
 * 
 * It is important that just touching is not enough; the two rooms have to
 * share at least three points of wall for there to be a door that is not in a
 * corner.
 * 
 * I have used method 2. This favors tightly packed dungeons, which I think is
 * nice.
 ******************************************************************************
 * 
 * How to generate the size of a new room:
 * 
 * This is another topic not covered in SDGA. I have used the following mini
 * algorithm:
 * 
 * 1. Calculate the average width and height of a room in a dungeon with a
 * width and a height and where a certain amount of space is to be left empty
 * at the end of the map generation.
 * 
 * 2. Set the max width to {@code (2 * average width - 3)} and set the max
 * height to {@code (2 * average height - 3)}. If {@code max width < 3}, set it
 * to 3, and the same for max height.
 * 
 * 3. Generate room sizes by finding random numbers between 3 and max width,
 * and 3 and max height, respectively.
 * 
 * @author Niels G. W. Serup
 * @version 1.0
 */
public class DungeonMapGenerator
{
    public static final int MIN_1D_SIZE = 3;

    public static final int DEFAULT_DUNGEON_WIDTH = 40;
    public static final int DEFAULT_DUNGEON_HEIGHT = 30;
    public static final int DEFAULT_NUMBER_OF_ROOMS_THRESHOLD = 24;
    public static final int DEFAULT_SPACE_LEFT_THRESHOLD = 100;
    public static final int DEFAULT_ERROR_THRESHOLD = 10;
    public static final int DEFAULT_NUMBER_OF_MONSTERS = 10;
    public static final int DEFAULT_NUMBER_OF_ITEMS = 11;
    public static final int DEFAULT_SCALE_FACTOR = 20;

    private static final DungeonMapGenerator instance
        = new DungeonMapGenerator();
    private static final Direction[] directions
        = { Direction.NORTH, Direction.EAST, Direction.SOUTH, Direction.WEST };
    
    private int dungeonWidth;
    private int dungeonHeight;
    private int numberOfRoomsThreshold;
    private int spaceLeftThreshold;
    private int errorThreshold;
    private int numberOfMonsters;
    private int numberOfItems;

    private ArrayList<Rectangle> rects;
    private Dungeon dungeon;
    private int maxWidth;
    private int maxHeight;
    private int spaceLeft;
    private boolean errorThresholdReached;

    private class Dimension
    {
        public int width;
        public int height;
        public Dimension(int width, int height)
        {
            this.width = width;
            this.height = height;
        }
    }
    
    private DungeonMapGenerator()
    {
        
    }

    /*
     * Generates a dungeon and then returns its corresponding dungeon map. A
     * DungeonMap stores all data as flat as a .map file, while Dungeon uses
     * various techniques to keep a dungeon structured, which means it is
     * easier to work with a dungeon than with a dungeon map. This is why a
     * dungeon map is not simply created from the beginning.
     */
    private DungeonMap generate
        (int dungeonWidth, int dungeonHeight, int numberOfRoomsThreshold,
         int spaceLeftThreshold, int errorThreshold, int numberOfMonsters,
         int numberOfItems, int scaleFactor, String description)
    {
        this.dungeonWidth = dungeonWidth;
        this.dungeonHeight = dungeonHeight;
        this.numberOfRoomsThreshold = numberOfRoomsThreshold;
        this.spaceLeftThreshold = spaceLeftThreshold;
        this.errorThreshold = errorThreshold;
        this.numberOfMonsters = numberOfMonsters;
        this.numberOfItems = numberOfItems;
        this.rects = new ArrayList<Rectangle>();

        if (this.dungeonWidth < MIN_1D_SIZE
            || this.dungeonHeight < MIN_1D_SIZE)
            throw new MapTooSmall();

        double averageRoomSize = (((double) (this.dungeonWidth
                                             * this.dungeonHeight
                                           - this.spaceLeftThreshold))
                                 / this.numberOfRoomsThreshold);
        double ratio = (double) this.dungeonWidth / this.dungeonHeight;
        /*
         * A := the area (averageRoomSize)
         * w := the width
         * h := the height
         * r := the width-height ratio (ratio)
         * 
         * A = w * h
         * r = w / h
         * 
         * r * A = w * h * (w / h) = w * w  =>  w = sqrt(r * A)
         *                                      h = w / r
         */
        double averageWidth = Math.sqrt(ratio * averageRoomSize);
        double averageHeight = averageWidth / ratio;
        this.maxWidth = (int) (averageWidth * 2.0 - MIN_1D_SIZE);
        this.maxHeight = (int) (averageHeight * 2.0 - MIN_1D_SIZE);
        if (this.maxWidth < MIN_1D_SIZE)
            this.maxWidth = MIN_1D_SIZE;
        if (this.maxHeight < MIN_1D_SIZE)
            this.maxHeight = MIN_1D_SIZE;
        
        this.dungeon = new Dungeon(scaleFactor, description);
        this.addRoomsAndDoors();
        this.addMonstersAndItems();

        return new DungeonMap(this.dungeon);
    }

    private void addRoomsAndDoors()
    {
        this.spaceLeft = this.dungeonWidth * this.dungeonHeight;
        this.errorThresholdReached = false;

        this.addFirstRoom();

        while (this.rects.size() < this.numberOfRoomsThreshold
               && this.spaceLeft > this.spaceLeftThreshold
               && !this.errorThresholdReached)
            {
                this.addSubsequentRoom();
            }
    }

    private void addRoom(Rectangle rect)
    {
        this.rects.add(rect);
        this.dungeon.addRoom(new Room(dungeon, rect));
        this.spaceLeft -= rect.getSize();
    }

    private void addFirstRoom()
    {
        Dimension first = this.createRectSize();
        int upperLeftX = randint(0, this.dungeonWidth - first.width);
        int upperLeftY = randint(0, this.dungeonHeight - first.height);
        Rectangle rect = new Rectangle(new Point(upperLeftX, upperLeftY),
                                       new Point(upperLeftX + first.width,
                                                 upperLeftY + first.height));
        this.addRoom(rect);
    }

    private void addSubsequentRoom()
    {
        Dimension rectSize = this.createRectSize();
        int errorsLeft = this.errorThreshold;
        while (errorsLeft >= 0 && rectSize.height >= MIN_1D_SIZE
               && rectSize.width >= MIN_1D_SIZE)
            {
                if (!this.tryPlaceRect(rectSize))
                    {
                        rectSize = this.createSmallerRectSize(rectSize);
                        errorsLeft--;
                    }
                else
                    return;
            }
        this.errorThresholdReached = true;
    }

    private boolean tryPlaceRect(Dimension rectSize)
    {
        int startTouchRectIndex = new Random().nextInt(this.rects.size());
        for (int i = startTouchRectIndex, j = 0;
             j < this.rects.size();
             i = (i + 1) % this.rects.size(), j++)
            {
                if (this.tryPlaceRectTouching(this.rects.get(i), rectSize))
                    return true;
            }
        return false;
    }

    // Tries to place a room of a given size whose boundary touches a given
    // rectangle
    private boolean tryPlaceRectTouching(Rectangle touch, Dimension rectSize)
    {
        for (int i = 0; i < 4; i++)
            {
                if (tryPlaceRectTouchingDirection
                    (touch, rectSize, DungeonMapGenerator.directions[i]))
                    return true;
            }
        return false;
    }

    private boolean tryPlaceRectTouchingDirection
        (Rectangle touch, Dimension rectSize, Direction direction)
    {
        if (direction == Direction.EAST || direction == Direction.WEST)
            {
                int x;
                if (direction == Direction.EAST)
                    {
                        x = touch.getLowerRight().getX();
                        if (x + rectSize.width > this.dungeonWidth)
                            return false;
                    }
                else
                    {
                        x = touch.getUpperLeft().getX() - rectSize.width;
                        if (x < 0)
                            return false;
                    }
                int startY = touch.getUpperLeft().getY() - (rectSize.height
                                                            - MIN_1D_SIZE);
                if (startY < 0)
                    startY = 0;
                int endY = touch.getLowerRight().getY() - MIN_1D_SIZE;
                if (endY + rectSize.height > this.dungeonHeight)
                    endY -= endY + rectSize.height - this.dungeonHeight;

                int firstTryY = randint(startY, endY);
                for (int y = firstTryY, j = 0; j <= endY - startY; j++)
                    {
                        if (this.tryPlaceActualRect
                            (touch, direction,
                             new Rectangle(new Point(x, y),
                                           new Point(x + rectSize.width,
                                                     y + rectSize.height))))
                            {
                                return true;
                            }
                        y++;
                        if (y > endY)
                            y = startY;
                    }
            }
        else if (direction == Direction.SOUTH || direction == Direction.NORTH)
            {
                int y;
                if (direction == Direction.SOUTH)
                    {
                        y = touch.getLowerRight().getY();
                        if (y + rectSize.height > this.dungeonHeight)
                            return false;
                    }
                else
                    {
                        y = touch.getUpperLeft().getY() - rectSize.height;
                        if (y < 0)
                            return false;
                    }
                int startX = touch.getUpperLeft().getX() - (rectSize.width
                                                            - MIN_1D_SIZE);
                if (startX < 0)
                    startX = 0;
                int endX = touch.getLowerRight().getX() - MIN_1D_SIZE;
                if (endX + rectSize.width > this.dungeonWidth)
                    endX -= endX + rectSize.width - this.dungeonWidth;
                int firstTryX = randint(startX, endX);
                for (int x = firstTryX, j = 0; j <= endX - startX; j++)
                    {
                        if (this.tryPlaceActualRect
                            (touch, direction,
                             new Rectangle(new Point(x, y),
                                           new Point(x + rectSize.width,
                                                     y + rectSize.height))))
                            {
                                return true;
                            }
                        x++;
                        if (x > endX)
                            x = startX;
                    }
            }
        return false;
    }

    private boolean tryPlaceActualRect
        (Rectangle touchRect, Direction direction, Rectangle newRect)
    {
        for (Rectangle rect : this.rects)
            {
                if (rect.overlaps(newRect))
                    return false;
            }
        this.addRoom(newRect);
        
        this.addDoor(touchRect, newRect, direction);
        for (int i = this.rects.size() - 2; i >= 0; i--)
            {
                if (this.rects.get(i) != touchRect
                    && new Random().nextBoolean()
                    && new Random().nextBoolean())
                    {
                        this.addDoor(this.rects.get(i), newRect);
                    }
            }
        return true;
    }

    // Add a two-way door to two rooms if possible
    private void addDoor(Rectangle rect1, Rectangle rect2)
    {
        this.addDoor(rect1, rect2, rect1.touches(rect2, MIN_1D_SIZE));
    }

    // Add a two-way door to two rooms that you are sure touches each other
    // with a specific direction.
    private void addDoor(Rectangle rect1, Rectangle rect2,
                         Direction touchDirection)
    {
        if (touchDirection == Direction.EAST
            || touchDirection == Direction.WEST)
            {
                int rect1X;
                if (touchDirection == Direction.EAST)
                    rect1X = rect1.getLowerRight().getX() - 1;
                else
                    rect1X = rect2.getLowerRight().getX() - 1;
                int rect2X = rect1X + 1;
                int rect1StartY = rect1.getUpperLeft().getY() + 1;
                int rect2StartY = rect2.getUpperLeft().getY() + 1;
                int startY = (rect1StartY > rect2StartY ? rect1StartY
                              : rect2StartY);
                int rect1EndY = rect1.getLowerRight().getY() - 1;
                int rect2EndY = rect2.getLowerRight().getY() - 1;
                int endY = (rect1EndY < rect2EndY ? rect1EndY : rect2EndY);
                int y = randrange(startY, endY);
                this.dungeon.addDoor(new Point(rect1X, y),
                                     new Point(rect2X, y));
            }
        else if (touchDirection == Direction.SOUTH
                 || touchDirection == Direction.NORTH)
            {
                int rect1Y;
                if (touchDirection == Direction.SOUTH)
                    rect1Y = rect1.getLowerRight().getY() - 1;
                else
                    rect1Y = rect2.getLowerRight().getY() - 1;
                int rect2Y = rect1Y + 1;
                int rect1StartX = rect1.getUpperLeft().getX() + 1;
                int rect2StartX = rect2.getUpperLeft().getX() + 1;
                int startX = (rect1StartX > rect2StartX ? rect1StartX
                              : rect2StartX);
                int rect1EndX = rect1.getLowerRight().getX() - 1;
                int rect2EndX = rect2.getLowerRight().getX() - 1;
                int endX = (rect1EndX < rect2EndX ? rect1EndX : rect2EndX);
                int x = randrange(startX, endX);
                this.dungeon.addDoor(new Point(x, rect1Y),
                                     new Point(x, rect2Y));
            }
    }

    private Dimension createRectSize()
    {
        int width = randint(MIN_1D_SIZE, this.maxWidth);
        int height = randint(MIN_1D_SIZE, this.maxHeight);
        return new Dimension(width, height);
    }

    private Dimension createSmallerRectSize(Dimension dim)
    {
        int newWidth = dim.width - 1;
        int newHeight = dim.height - 1;
        return new Dimension(newWidth, newHeight);
    }

    private void addMonstersAndItems()
    {
        int i;
        for (i = 0; i < this.numberOfMonsters; i++)
            {
                Monster monster = (new Random().nextBoolean()
                                   ? new Orc(this.dungeon)
                                   : new Goblin(this.dungeon));
                monster.placeAtRandomEmptyPosition();
                this.dungeon.addMonster(monster);
            }
        for (i = 0; i < this.numberOfItems; i++)
            {
                Item item = (new Random().nextBoolean()
                             ? new Weapon(this.dungeon, randint(10, 75))
                             : new Armor(this.dungeon, randint(1, 99)));
                item.placeAtRandomEmptyPosition();
                this.dungeon.addItem(item);
            }
    }

    /**
     * Generates a dungeon map according to the rules laid out, except it might
     * in rare cases create only one room with no doors (and all rooms should
     * have doors). If any of the integers are null, replace them with default
     * values.
     * 
     * @throws MapTooSmall  if not even one room can be created
     * @return the new dungeon map
     */
    public static DungeonMap generateDungeonMap
        (Integer dungeonWidth, Integer dungeonHeight,
         Integer numberOfRoomsThreshold, Integer spaceLeftThreshold,
         Integer errorThreshold, Integer numberOfMonsters,
         Integer numberOfItems, Integer scaleFactor, String description)
    {
        if (dungeonWidth == null)
            dungeonWidth = DungeonMapGenerator.DEFAULT_DUNGEON_WIDTH;
        if (dungeonHeight == null)
            dungeonHeight = DungeonMapGenerator.DEFAULT_DUNGEON_HEIGHT;
        if (numberOfRoomsThreshold == null)
            numberOfRoomsThreshold = DungeonMapGenerator
                .DEFAULT_NUMBER_OF_ROOMS_THRESHOLD;
        if (spaceLeftThreshold == null)
            spaceLeftThreshold = DungeonMapGenerator
                .DEFAULT_SPACE_LEFT_THRESHOLD;
        if (errorThreshold == null)
            errorThreshold = DungeonMapGenerator.DEFAULT_ERROR_THRESHOLD;
        if (numberOfMonsters == null)
            numberOfMonsters = DungeonMapGenerator.DEFAULT_NUMBER_OF_MONSTERS;
        if (numberOfItems == null)
            numberOfItems = DungeonMapGenerator.DEFAULT_NUMBER_OF_ITEMS;
        if (scaleFactor == null)
            scaleFactor = DungeonMapGenerator.DEFAULT_SCALE_FACTOR;

        return DungeonMapGenerator.instance.generate
            (dungeonWidth, dungeonHeight, numberOfRoomsThreshold,
             spaceLeftThreshold, errorThreshold, numberOfMonsters,
             numberOfItems, scaleFactor, description);
    }

    /**
     * Generates a default dungeon map.
     */
    public static DungeonMap generateDungeonMap()
    {
        return DungeonMapGenerator.generateDungeonMap
            (null, null, null, null, null, null, null, null, null);
    }

    private static int randrange(int a, int b)
    {
        return new Random().nextInt(b - a) + a;
    }

    private static int randint(int a, int b)
    {
        return randrange(a, b + 1);
    }
}
