/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.model;
import dungeoncrawl.general.*;
import dungeoncrawl.model.*;
import dungeoncrawl.model.item.*;

import java.util.ArrayList;
import java.util.Random;

/**
 * A dungeon object, superclass to both Player and Item. All dungeon objects
 * know of their own positions (if they have positions) and the dungeon to
 * which they are associated. This might seem to complicate things, as first
 * the dungeon has to "be added" to the dungeon object, and then the dungeon
 * object has to "be added" to the dungeon, but since this game is based on the
 * simple rule that at any time there is only one playable dungeon, this is not
 * a problem. It actually makes certain things much easier, because the object
 * knows a lot about itself, itself. It does not have to ask the dungeon for
 * help.
 *
 * Something which is by definition not modifiable (such as a room or a door)
 * should not be a subclass of DungeonObject, because DungeonObject is designed
 * for modifiable things.
 * 
 * @author Niels G. W. Serup
 * @version 1.0
 */
public abstract class DungeonObject
{
    private Dungeon dungeon;
    private Point position;

    /**
     * Creates an object associated with {@code dungeon}. {@code position} can
     * be null, but remember to set it later with {@code setPosition}.
     * @require dungeon != null
     */
    public DungeonObject(Dungeon dungeon, Point position)
    {
        if (dungeon == null)
            {
                throw new NullPointerException();
            }
        this.dungeon = dungeon;
        this.position = position;
    }

    /**
     * Creates an object associated with {@code dungeon} and with no position.
     * @require dungeon != null
     */
    public DungeonObject(Dungeon dungeon)
    {
        this(dungeon, null);
    }

    /**
     * Sets the position of the object.
     */
    public void setPosition(Point position)
    {
        this.position = position;
    }

    /**
     * @return the position of the object
     */
    public Point getPosition()
    {
        return this.position;
    }

    /**
     * @return the dungeon of the object
     */
    public Dungeon getDungeon()
    {
        return this.dungeon;
    }

    /**
     * Places the object at a random position.
     */
    public void placeAtRandomEmptyPosition()
    {
        Point randomPosition = this.getRandomEmptyPosition();
        if (randomPosition == null)
            {
                throw new NoRandomPosition();
            }
        this.setPosition(randomPosition);
    }

    /**
     * Tries to find a point in all rooms where {@code this.isPositionGood}
     * returns true. A door is considered an empty position.
     * @return such a point if it exists; otherwise {@code null}
     */
    public Point getRandomEmptyPosition()
    {
        Random random = new Random();
        ArrayList<Room> rooms = this.getDungeon().getRooms();
        int startRoom = random.nextInt(rooms.size());
        Point tempPoint;
        for (int roomI = startRoom, i = 0;
             i < rooms.size();
             roomI = (roomI + 1) % rooms.size(), i++)
            {
                tempPoint = rooms.get(roomI).getRandomEmptyPosition(this);
                if (tempPoint != null)
                    return tempPoint;
            }
        return null;
    }

    /**
     * @return whether or not a position is ok for placement of this dungeon
     * object
     */
    public abstract boolean isPositionGood(Point position);
}
