/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.model;
import dungeoncrawl.general.*;
import dungeoncrawl.model.player.monster.*;
import dungeoncrawl.model.item.*;


import java.io.File;
import java.io.FileNotFoundException;
import java.util.HashMap;
import java.util.Scanner;
import java.util.NoSuchElementException;
import java.util.regex.Pattern;

/**
 * A parser for the special dungeon map language.
 * 
 * Map format:
 * <pre>
 * [{@literal !<description>}]
 * {@literal <scale>}
 * room: {@literal <x1> <y1> <x2> <y2>}
 *          :
 *          :
 * door: {@literal <x1> <y1> <x2> <y2>}
 *          :
 *          :
 * monster: {@literal <x> <y>} O|G
 *          :
 *          :
 * item: {@literal <x> <y>} A|W <strength>
 *          :
 *          :
 * </pre>
 * where O = Orc and G = Goblin.
 * @author Niels G. W. Serup
 * @version 1.0
 */
public class DungeonParser
{
    enum Action
    {
        ROOM
            {
                public void execute(DungeonParser parser)
                {
                    int x1 = parser.scanner.nextInt();
                    int y1 = parser.scanner.nextInt();
                    int x2 = parser.scanner.nextInt();
                    int y2 = parser.scanner.nextInt();
                    parser.dungeonMap.addRoom(new Point(x1, y1),
                                              new Point(x2, y2));
                }
            },

        DOOR
            {
                public void execute(DungeonParser parser)
                {
                    int x1 = parser.scanner.nextInt();
                    int y1 = parser.scanner.nextInt();
                    int x2 = parser.scanner.nextInt();
                    int y2 = parser.scanner.nextInt();
                    parser.dungeonMap.addDoor(new Point(x1, y1),
                                              new Point(x2, y2));
                }
            },

        MONSTER
            {
                public void execute(DungeonParser parser)
                {
                    int x = parser.scanner.nextInt();
                    int y = parser.scanner.nextInt();
                    String type = parser.scanner.next().toUpperCase();
                    Point p = new Point(x, y);
                    if (type.equals("O"))
                        {
                            parser.dungeonMap.addOrc(p);
                        }
                    else if (type.equals("G"))
                        {
                            parser.dungeonMap.addGoblin(p);
                        }
                }
            },

        ITEM
            {
                public void execute(DungeonParser parser)
                {
                    int x = parser.scanner.nextInt();
                    int y = parser.scanner.nextInt();
                    String type = parser.scanner.next();
                    int strength = parser.scanner.nextInt();
                    Point p = new Point(x, y);
                    if (type.equals("A"))
                        {
                            parser.dungeonMap.addArmor(p, strength);
                        }
                    else if (type.equals("W"))
                        {
                            parser.dungeonMap.addWeapon(p, strength);
                        }
                }
            };

        public abstract void execute(DungeonParser parser);
    }

    private static final Pattern DEFAULT_PATTERN = Pattern.compile("\\s+");
    private static final Pattern TYPE_PATTERN = Pattern.compile("[\\s:]+");
    private static DungeonParser instance = new DungeonParser();
    private Scanner scanner;
    private DungeonMap dungeonMap;
    private HashMap<String, Action> actions;

    private DungeonParser()
    {
        this.actions = new HashMap<String, Action>();
        this.actions.put("room", Action.ROOM);
        this.actions.put("door", Action.DOOR);
        this.actions.put("monster", Action.MONSTER);
        this.actions.put("item", Action.ITEM);
        this.scanner = null;
        this.dungeonMap = null;
    }
    
    private DungeonMap parseM(Scanner scanner)
    {
        this.scanner = scanner;
        int scaleFactor;
        String description = this.scanner.nextLine().trim();
        if (description.startsWith("!"))
            {
                description = description.substring(1).trim();
                scaleFactor = this.scanner.nextInt();
            }
        else
            {
                scaleFactor = Integer.parseInt(description);
                description = null;
            }
        this.dungeonMap = new DungeonMap(scaleFactor, description);
        String type;
        while (true)
            {
                try
                    {
                        this.scanner.useDelimiter(DungeonParser.TYPE_PATTERN);
                        type = this.scanner.next();
                        this.actions.get(type).execute(this);
                        this.scanner.useDelimiter(DungeonParser.DEFAULT_PATTERN);
                    }
                catch (NoSuchElementException e)
                    {
                        break;
                    }
            }
        return this.dungeonMap;
    }

    /**
     * Parses a dungeon map. Assumes the file information is valid.
     * @return the new Dungeon map
     */
    public static DungeonMap parseMap(Scanner scanner)
    {
        return instance.parseM(scanner);
    }

    /**
     * Parses "~/dungeon.map".
     * @return the new Dungeon
     */
    public static DungeonMap parseMap() throws FileNotFoundException
    {
        return DungeonParser.instance.parseM
            (new Scanner(new File(new File(System.getProperty("user.home")),
                                  "dungeon.map")));
    }
}