/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.model;
import dungeoncrawl.general.*;
import dungeoncrawl.model.player.Player;
import java.util.Collection;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.Random;

/**
 * A room. Each room controls its own doors. Doors, like rooms, cannot be
 * modified once they have been created. They are absolute.
 * 
 * NOTE: While it may have been the intent of the E.1 that a lot of things
 * should be stored in rooms and not the dungeon, I have not done this. A room
 * controls only the doors within it. Everything else is stored in
 * Dungeon. This makes especially good sense for players, as they can move to
 * and from rooms. One could have chosen to store items in rooms and not the
 * dungeon, but it makes it easier to extend the game if they are stored in a
 * Dungeon instance, as in the future one might want to implement the ability
 * to drop items.
 * 
 * @author Niels G. W. Serup
 * @version 1.0
 */
public class Room
{
    private HashMap<Point, Door> doors;
    private Rectangle roomBoundary;
    private Dungeon dungeon;
    
    /**
     * Creates a room associated with a dungeon from a rectangle.
     * @require roomBoundary != null
     * @require dungeon != null
     */
    public Room(Dungeon dungeon, Rectangle roomBoundary)
    {
        if (roomBoundary == null || dungeon == null)
            {
                throw new NullPointerException();
            }
        this.roomBoundary = roomBoundary;
        this.dungeon = dungeon;
        this.doors = new HashMap<Point, Door>();
    }

    /**
     * Creates a room from two points.
     */
    public Room(Dungeon dungeon, Point upperLeft, Point lowerRight)
    {
        this(dungeon, new Rectangle(upperLeft, lowerRight));
    }

    public Rectangle getBoundary()
    {
        return this.roomBoundary;
    }

    public Dungeon getDungeon()
    {
        return this.dungeon;
    }

    public Door getDoorAt(Point position)
    {
        return this.doors.get(position);
    }

    /**
     * @return the doors which connect this room to another given room. It is
     * ok to modify the returned list.
     */
    public ArrayList<Door> getConnections(Room otherRoom)
    {
        ArrayList<Door> connections = new ArrayList<Door>();
        for (Door door : this.doors.values())
            {
                if (door.getTargetRoom() == otherRoom)
                    connections.add(door);
            }
        return connections;
    }

    /**
     * @return the doors of the room. It is not ok to modify the returned
     * collection.
     */
    public Collection<Door> getDoors()
    {
        return this.doors.values();
    }

    /**
     * Tries to find a point in the room with no monster and no item. A door is
     * considered an empty position.
     * @return such a point if it exists; otherwise {@code null}
     */
    public Point getRandomEmptyPosition(DungeonObject dungeonObject)
    {
        Random random = new Random();
        int height = this.roomBoundary.getHeight();
        int width = this.roomBoundary.getWidth();
        int numOfPoints = this.roomBoundary.getSize();
        int start1DPosition = random.nextInt(numOfPoints);
        Point tempPoint;
        for (int position = start1DPosition, i = 0;
             i < numOfPoints;
             position = (position + 1) % numOfPoints, i++)
            {
                tempPoint = new Point
                    (position % width
                     + this.getBoundary().getUpperLeft().getX(),
                     position / width
                     + this.getBoundary().getUpperLeft().getY());
                if (dungeonObject.isPositionGood(tempPoint))
                    {
                        return tempPoint;
                    }
            }
        return null;
    }

    /**
     * @return {@code true} if this room and another given room are connected
     * through a door; {@code false} otherwise
     */
    public boolean isConnected(Room otherRoom)
    {
        for (Door door : this.getDoors())
            {
                if (door.getTargetRoom() == otherRoom)
                    return true;
            }
        return false;
    }

    /**
     * @return {@code true} if there is a path from this room to another room
     * through 0 or more other rooms; {@code false} otherwise
     */
    public boolean hasPath(Room otherRoom)
    {
        LinkedList<Room> visitedRooms = new LinkedList<Room>();
        return this.hasPath(otherRoom, visitedRooms);
    }

    private boolean hasPath(Room otherRoom, LinkedList<Room> visitedRooms)
    {
        visitedRooms.add(this);
        Room room;
        for (Door door : this.getDoors())
            {
                room = door.getTargetRoom();
                if (room == otherRoom
                    || (!visitedRooms.contains(room)
                        && room.hasPath(otherRoom, visitedRooms)))
                    return true;
            }
        return false;
    }

    /**
     * @return {@code true} if a given position is inside the room; {@code
     * false} otherwise.
     */
    public boolean isInside(Point position)
    {
        return this.roomBoundary.isInside(position);
    }

    /**
     * Adds a one-way door to whatever room is behind the wall. Assumes the
     * door is at a wall but not in a corner.
     */
    public void addDoor(Point position, Point targetPosition, Room targetRoom)
    {
        if (position != null)
            {
                this.doors.put(position, new Door(position, this,
                                                  targetPosition, targetRoom));
            }
    }

    /**
     * Removes the given player from the room if present.
     * 
     * This method is deprecated. Use {@code Dungeon.removePlayer} instead.
     * @deprecated
     */
    public void removePlayer(Player player)
    {
        if (this.isInside(player.getPosition()))
            this.dungeon.removePlayer(player);
    }
}