/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.model.movementstrategy;
import dungeoncrawl.general.*;
import dungeoncrawl.model.player.monster.Monster;
import dungeoncrawl.model.*;

import java.util.Collection;
import java.util.ArrayList;
import java.util.Random;

/**
 * Move towards the shortest Manhattan distance, but use doors.
 * 
 * This class does not override the {@code getTargetApproximity(Point target)}
 * method, since it still uses the Manhattan approach.
 * @author Niels G. W. Serup
 * @version 1.0
 */
public class BetterMovementStrategy extends ManhattanMovementStrategy
{
    private Door previousDoor;
    private Door randomDoor;

    public BetterMovementStrategy()
    {
        super();
    }
    
    public BetterMovementStrategy(Monster monster)
    {
        super(monster);
    }
    
    /**
     * Improves upon ManhattanMovementStrategy by letting a monster walk
     * towards doors if the hero is not in the same room as the monster.
     *
     * More specifically, these three rules are used:
     * 
     * If the hero is in the same room as the monster, the monster moves
     * towards the hero the same way as in ManhattanMovementStrategy.
     * 
     * If the hero is in another room which has a door to the monster's room,
     * the monster moves towards that door, and if it reaches it, moves through
     * it.
     * 
     * If the hero is in a faraway room, the monster chooses a door in its room
     * and moves towards that door (with the same intent of eventually moving
     * through it). The door must not be the previously moven through door, if
     * the monster has moved through any doors, unless there is only one door
     * in the room.
     */
    @Override
    public void move(Point target)
    {
        Point monsterPos = this.monster.getPosition();
        Room monsterRoom = this.dungeon.getRoomOfPosition(monsterPos);
        Room targetRoom = this.dungeon.getRoomOfPosition(target);
        if (monsterRoom == targetRoom)
            {
                super.move(target);
                this.randomDoor = null;
                return;
            }
        if (this.moveTowardsClosestDoor(monsterRoom, targetRoom))
            return;
        this.moveTowardsRandomDoor(monsterRoom, targetRoom);
    }

    // Returns true if the monster has entered a new room; false otherwise.
    private boolean moveAndRemember(Door targetDoor)
    {
        Point monsterPos = this.monster.getPosition();
        if (monsterPos.equals(targetDoor.getPosition()))
            {
                return this.tryMoveThroughDoorAndRemember(targetDoor);
            }
        else
            {
                super.move(targetDoor.getPosition());
                return false;
            }
    }

    private boolean tryMoveThroughDoorAndRemember(Door targetDoor)
    {
        super.move(targetDoor.getTargetPosition());
        if (this.monster.getPosition().equals(targetDoor.getTargetPosition()))
            {
                this.previousDoor = targetDoor.getTargetRoom().getDoorAt
                    (targetDoor.getTargetPosition());
                return true;
            }
        return false;
    }

    // Returns true if succesful; false otherwise
    private boolean moveTowardsClosestDoor(Room monsterRoom, Room targetRoom)
    {
        Door closestDoor = this.getClosestConnection(monsterRoom, targetRoom);
        if (closestDoor != null)
            {
                this.moveAndRemember(closestDoor);
                this.randomDoor = null;
                return true;
            }
        return false;
    }

    // Returns the closest connection (door) or null if none exist
    private Door getClosestConnection(Room monsterRoom, Room targetRoom)
    {
        ArrayList<Door> connections = monsterRoom.getConnections(targetRoom);
        if (connections.size() == 0)
            return null;
        Door closestDoor = connections.get(0);
        int minDist = this.getManhattanDistance(closestDoor.getPosition());
        connections.remove(0);
        int tempDist;
        for (Door door : connections)
            {
                tempDist = this.getManhattanDistance(door.getPosition());
                if (tempDist > minDist)
                    {
                        closestDoor = door;
                        minDist = tempDist;
                    }
            }
        return closestDoor;
    }

    // If no random door path (this.randomDoor) has been set, sets a new random
    // path towards a door.
    private void moveTowardsRandomDoor(Room monsterRoom, Room targetRoom)
    {
        Door targetDoor;
        if (this.randomDoor != null)
            {
                targetDoor = this.randomDoor;
            }
        else
            {
                targetDoor = this.getRandomDoor(monsterRoom, targetRoom);
                this.randomDoor = targetDoor;
            }
        if (targetDoor == null)
            return;
        if (this.moveAndRemember(targetDoor))
            this.randomDoor = null;
    }

    // Returns a random door which is not the previously exited door, except if
    // that is the only door in monsterRoom.
    private Door getRandomDoor(Room monsterRoom, Room targetRoom)
    {
        Collection<Door> doors = monsterRoom.getDoors();
        if (doors.size() == 0)
            return null;
        if (doors.size() == 1)
            {
                for (Door door : doors)
                    return door;
            }
        
        Door targetDoor;
        int i = new Random().nextInt(this.previousDoor == null
                                     ? doors.size() : doors.size() - 1);
        int j = 0;
        for (Door door : doors)
            {
                if (j == i)
                    {
                        if (door == this.previousDoor)
                            i++;
                        else
                            return door;
                    }
                j++;
            }
        return null; // code path will never reach here
    }
}
