/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.model.movementstrategy;
import dungeoncrawl.general.*;
import dungeoncrawl.model.*;
import dungeoncrawl.model.player.monster.Monster;

/**
 * Movement strategies for monsters.
 * @author Niels G. W. Serup
 * @version 1.0
 */
public abstract class MovementStrategy
{
    protected Monster monster;
    protected Dungeon dungeon;

    /**
     * Creates an inactive movement strategy.
     * 
     * You must remember to set its monster via {@code setMonster(Monster
     * monster)} later on if you use this constructor.
     */
    public MovementStrategy()
    {

    }

    /**
     * Creates an active movement strategy.
     */
    public MovementStrategy(Monster monster)
    {
        this.setMonster(monster);
    }

    /**
     * @assume monster != null
     */
    public void setMonster(Monster monster)
    {
        this.monster = monster;
        this.dungeon = monster.getDungeon();
    }
    
    /**
     * Moves a monster towards a target position.
     */
    public abstract void move(Point target);

    /**
     * This is not to get a precise distance, taking into account what a
     * specific movement strategy normally takes into account (i.e. one
     * movement strategy might know about doors), but to get a circa distance,
     * with the sole aim of moving monsters in a way that allows for most
     * monsters to move.
     * 
     * This does not solve everything; monsters might still end in deadlocks if
     * a more complex movement strategy is used.
     * @see Dungeon.moveMonstersAutonomously
     * 
     * @return how close the movement strategy thinks its monster is to the
     * target
     */
    public abstract int getTargetApproximity(Point target);

    /**
     * Moves a monster towards the hero.
     */
    public void moveTowardsHero()
    {
        this.move(this.dungeon.getHero().getPosition());
    }
    
    /**
     * @return how close the movement strategy thinks its monster is to the
     * hero
     */
    public int getHeroApproximity()
    {
        return this.getTargetApproximity(this.dungeon.getHero().getPosition());
    }
}
