/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.model.player;
import dungeoncrawl.general.*;
import dungeoncrawl.model.*;
import dungeoncrawl.model.item.*;
import dungeoncrawl.notification.PlayerMoved;


/**
 * A player, superclass to both Hero and Monster.
 * @author Niels G. W. Serup
 * @version 1.0
 */
public abstract class Player extends DungeonObject
{
    public final static int MAX_HEALTH = 100;
    private int health;
    private int regenerationLevel;

    /**
     * Creates a player associated with {@code dungeon} and with the initial
     * health of {@code Player.MAX_HEALTH}. {@code position} can be null, but
     * remember to set it later with {@code setPosition}.
     * @require dungeon != null
     */
    public Player(Dungeon dungeon, Point position)
    {
        super(dungeon, position);
        this.health = Player.MAX_HEALTH;
    }

    /**
     * Creates a player associated with {@code dungeon} and with the initial
     * health of {@code Player.MAX_HEALTH} and with no position. 
     * @require dungeon != null
     */
    public Player(Dungeon dungeon)
    {
        this(dungeon, null);
    }

    protected void setHealth(int health)
    {
        this.health = health;
    }

    protected void setRegenerationLevel(int regenerationLevel)
    {
        this.regenerationLevel = regenerationLevel;
    }
    /**
     * @return {@code true} if player’s health > 0; false otherwise.
     */
    public boolean isAlive()
    {
        return this.health > 0;
    }

    /**
     * @return {@code true} if player is not alive; false otherwise.
     */
    public boolean isDead()
    {
        return (!this.isAlive());
    }

    public int getHealth()
    {
        return this.health;
    }

    /**
     * Performs an amount of regeneration for a single turn.
     *
     * A player can at most regenerate to health level Player.MAX_HEALTH.  A dead
     * player cannot regenerate.
     * @ensure this.health <= Player.MAX_HEALTH
     * @ensure this.isDead()@pre && this.health <= 0
     */
    public void regenerate()
    {
        if (this.isAlive())
            {
                int newHealth = this.health + this.regenerationLevel;
                if (newHealth > Player.MAX_HEALTH)
                    {
                        this.health = Player.MAX_HEALTH;
                    }
                else
                    {
                        this.health = newHealth;
                    }
            }
    }

    /**
     * Checks if a move in some direction is considered ok.
     *
     * Does not check if the move can actually succeed and does not mutate
     * anything.
     * 
     * (Note: In actuality, only the hero needs this method, as the monsters
     * are not controllable. However, should one choose to extend Dungeon Crawl
     * so that one would also be able to control monsters, this would be
     * useful. As it is right now, it is both useless and harmless.)
     * @return whether or not the move is acceptable
     */
    public boolean isAcceptableMove(Direction direction)
    {
        return true;
    }

    /**
     * @return whether or not it is ok to hit {@code otherPlayer}. Defaults to
     * true.
     */
    public boolean mayHitOtherPlayer(Player otherPlayer)
    {
        return true;
    }

    /**
     * Attempts to move the player one point in the desired direction.
     * 
     * The move fails if there is a wall in the way.  If there is a player in
     * the target position, that player is damaged.  If that player dies due to
     * the damage, the move succeeds; otherwise it does not.
     * @assume this.position != null
     */
    public void tryMove(Direction direction)
    {
        Point newPosition = direction.getNewPosition(this.getPosition());
        if (this.getDungeon().isWallBetween(this.getPosition(), newPosition))
            return;
        Player otherPlayer = this.getDungeon().getPlayerAt(newPosition);
        if (otherPlayer == null)
            {
                this.moved(newPosition);
                return;
            }
        if (!this.mayHitOtherPlayer(otherPlayer))
            {
                return;
            }
        otherPlayer.takeDamage(this.getDamageLevel());
        if (otherPlayer.isDead())
            {
                this.moved(newPosition);
            }
    }

    /* 
     * What to do when a move has occured.
     */
    protected void moved(Point newPosition)
    {
        this.setPosition(newPosition);
        this.getDungeon().notifyObservers(new PlayerMoved(this));
    }

    /**
     * @ensure return >= 0
     * @return the level of damage that the player can incur on another player.
     */
    public abstract int getDamageLevel();

    /**
     * Damages the player with the given amount.
     */
    public abstract void takeDamage(int amount);
}
