/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.model.player.hero;
import dungeoncrawl.general.*;
import dungeoncrawl.model.player.Player;
import dungeoncrawl.model.player.monster.Monster;
import dungeoncrawl.model.item.*;
import dungeoncrawl.model.*;

import java.util.LinkedList;

/**
 * A generic hero.
 * 
 * A hero is meant to be moved by the human player, and so it does not require
 * any movement strategies.
 * 
 * If a hero walks over an item, he picks it up. If it is a piece of armor, he
 * adds it to his existing armor. If it is a weapon, he swaps it with his
 * current weapon if the new weapon is more powerful, or he throws it away (out
 * of the dungeon) if it is less powerful. The inventory does not keep track of
 * weapons that were less powerful than the hero's current weapon at the time
 * of discovery and were thrown away at once.
 * 
 * @author Niels G. W. Serup
 * @version 1.0
 */
public abstract class Hero extends Player
{
    private final static int BARE_HANDS_DAMAGE = 5;

    private int damageMagnitude;
    private String name;
    private Weapon currentWeapon;
    private Armor armor;
    private LinkedList<Item> inventory;

    /**
     * Creates a hero associated with {@code dungeon} and with no weapon or
     * armor but his/her bare fists.
     */
    public Hero(Dungeon dungeon, String name)
    {
        super(dungeon);
        this.name = name;
        this.currentWeapon = new Weapon(dungeon, Hero.BARE_HANDS_DAMAGE);
        this.armor = new Armor(dungeon);
        this.inventory = new LinkedList<Item>();
    }

    /**
     * Calculates the damage level from the damage level of the current weapon
     * and the damage magnitude of the hero.
     * @ensure return >= 0.
     * @return the level of damage that the hero can incur on another player.
     */
    @Override
    public int getDamageLevel()
    {
        return this.currentWeapon.getDamage() * this.damageMagnitude;
    }

    public String getName()
    {
        return this.name;
    }

    public LinkedList<Item> getInventory()
    {
        return this.inventory;
    }

    public Armor getArmor()
    {
        return this.armor;
    }

    public Weapon getWeapon()
    {
        return this.currentWeapon;
    }

    protected void setDamageMagnitude(int damageMagnitude)
    {
        this.damageMagnitude = damageMagnitude;
    }
    
    /**
     * Damages the hero's armor and eventually the hero with the given amount.
     */
    @Override
    public void takeDamage(int amount)
    {
        int armorResistance = this.armor.getResistance();
        if (amount <= armorResistance)
            {
                this.armor.decreaseResistance(amount);
            }
        else {
            this.armor.setResistance(0);
            this.setHealth(this.getHealth() - (amount - armorResistance));
        }
    }

    /**
     * Picks up armor.
     * @ensure this.armor.getResistance() >= this.armor.getResistance()@pre
     */
    public void pickUpArmor(Armor armor)
    {
        if (armor != null)
            {
                armor.setPosition(null);
                this.armor.increaseResistance(armor.getResistance());
                this.inventory.add(armor);
            }
    }

    /**
     * Picks up a weapon, compares its damage to the current weapon, and swaps
     * it if it is better, in which case the old weapon disappears.
     */
    public void pickUpWeapon(Weapon weapon)
    {
        if (weapon != null)
            {
                if (weapon.getDamage() > this.currentWeapon.getDamage())
                    {
                        weapon.setPosition(null);
                        this.currentWeapon = weapon;
                        this.inventory.add(weapon);
                    }
            }
    }

    /**
     * Checks if a wall is not in the way for a move in some direction.
     *
     * Does not check if the move can actually succeed and does not mutate
     * anything.
     * @assume this.getPosition() != null
     * @return whether or not the path is clear (not counting monsters and
     * items)
     */
    @Override
    public boolean isAcceptableMove(Direction direction)
    {
        Point newPosition = direction.getNewPosition(this.getPosition());
        return this.getDungeon().isClearPath(this.getPosition(), newPosition);
    }

    
    /* 
     * Also loots if standing on an item.
     */
    @Override
    protected void moved(Point newPosition)
    {
        super.moved(newPosition);
        Item itemAtPos = this.getDungeon().loot(newPosition);
        if (itemAtPos != null)
            {
                itemAtPos.getUsedByHero(this);
            }
    }

    /**
     * A hero cannot be placed on monsters nor items.
     */
    @Override
    public boolean isPositionGood(Point position)
    {
        return (this.getDungeon().getItemAt(position) == null
                && this.getDungeon().getPlayerAt(position) == null);
    }

    /**
     * A hero only attempts to move if he/she can, and after that the monsters
     * move.
     */
    @Override
    public void tryMove(Direction direction)
    {
        if (this.isAcceptableMove(direction))
            {
                super.tryMove(direction);
                this.getDungeon().afterHeroMove();
            }
    }
}
