/*
 * Dungeon Crawl of Doom
 * Copyright (C) 2012  Niels G. W. Serup

 * This file is part of Dungeon Crawl of Doom.

 * Dungeon Crawl of Doom is free software: you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.

 * Dungeon Crawl of Doom is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Enigma.  If not, see <http://www.gnu.org/licenses/>.
 */


package dungeoncrawl.model.player.monster;
import dungeoncrawl.general.*;
import dungeoncrawl.model.player.Player;
import dungeoncrawl.model.player.hero.Hero;
import dungeoncrawl.model.movementstrategy.*;
import dungeoncrawl.model.*;


/**
 * A generic monster.
 * 
 * A monster has to move on its own. To that end, it has movement
 * strategies. There are two: the simple "scent-based" strategy, and the more
 * advanced door-noticing strategy.
 * 
 * @see MovementStrategy
 * @see ManhattanMovementStrategy
 * @see BetterMovementStrategy
 * @author Niels G. W. Serup
 * @version 1.0
 */
public abstract class Monster extends Player
{
    private MovementStrategy movementStrategy;
    
    /**
     * Creates a Monster associated with {@code dungeon} put into a position
     * and with a specific move strategy. The monster associates itself with
     * the movement strategy, so it must not be used by other monsters. If
     * {@code movementStrategy} is null, use the default
     * BetterMovementStrategy.
     * @require dungeon != null
     */
    public Monster(Dungeon dungeon, Point position,
                   MovementStrategy movementStrategy)
    {
        super(dungeon, position);
        if (movementStrategy != null)
            this.setMovementStrategy(movementStrategy);
        else
            this.setMovementStrategy(new BetterMovementStrategy());
    }

    /**
     * Creates a Monster associated with {@code dungeon} put into a position
     * and with the default BetterMovementStrategy.
     * @require dungeon != null
     */
    public Monster(Dungeon dungeon, Point position)
    {
        this(dungeon, position, null);
    }

    /**
     * Creates a Monster associated with {@code dungeon} with no position and
     * with the default, dumb move strategy. Remember to set the position later
     * on.
     * @require dungeon != null
     */
    public Monster(Dungeon dungeon)
    {
        this(dungeon, null, null);
    }

    /**
     * Sets the move strategy.
     * @require movementStrategy != null
     */
    public void setMovementStrategy(MovementStrategy movementStrategy)
    {
        if (movementStrategy == null)
            {
                throw new NullPointerException();
            }
        this.movementStrategy = movementStrategy;
        this.movementStrategy.setMonster(this);
    }

    /**
     * Damages the monster's health.
     */
    @Override
    public void takeDamage(int amount)
    {
        this.setHealth(this.getHealth() - amount);
    }

    /**
     * @return {@code true} if {@code otherPlayer} is not a monster; {@code
     * false} otherwise.
     */
    public boolean mayHitOtherPlayer(Player otherPlayer)
    {
        return !(otherPlayer instanceof Monster);
    }

    /**
     * Moves the monster towards the hero through the use af a
     * MovementStrategy.
     */
    public void makeAutonomousMove()
    {
        this.movementStrategy.moveTowardsHero();
    }

    /**
     * @return the hero approximity based on the current movement strategy
     */
    public int getHeroApproximity()
    {
        return this.movementStrategy.getHeroApproximity();
    }

    /* 
     * Also updates the dungeon's knowledge of this monster's position.
     */
    @Override
    public void setPosition(Point newPosition)
    {
        this.getDungeon().updateMonsterPosition(this, newPosition);
        super.setPosition(newPosition);
    }

    /**
     * A monster cannot be placed on other players.
     */
    @Override
    public boolean isPositionGood(Point position)
    {
        return (this.getDungeon().getPlayerAt(position) == null);
    }
}
