// njf941

import dungeoncrawl.model.*;
import dungeoncrawl.general.*;

import static org.junit.Assert.*;
import org.junit.Test;

import java.util.Scanner;
import java.util.ArrayList;

/**
 * Testing of DungeonParser.
 * @author Niels G. W. Serup
 * @version 1.0
 */
public class DungeonParserTests {
    private static final String[] TEST_MAPS = new String[] {
        ("!This is the real map!\n"
         + "20\n"
         + "room: 0 0 4 5\n"
         + "room: 0 5 4 13\n"
         + "room: 4 1 7 8\n"
         + "room: 7 4 13 8\n"
         + "room: 4 8 13 13\n"
         + "door: 3 2 4 2\n"
         + "door: 3 6 4 6\n"
         + "door: 3 10 4 10\n"
         + "door: 11 7 11 8\n"
         + "monster: 10 5 O\n"
         + "monster: 10 8 G\n"
         + "monster: 5 9 G\n"
         + "item: 5 4 A 50\n"
         + "item: 5 5 W 10\n"
         + "item: 5 9 A 50\n"
         + "item: 6 11 W 20\n"
         + "item: 11 8 W 30\n"
         + "item: 11 5 A 50\n"
        ),
        ("20 \n"
         + "room:  0   0  4 5\n"
         + "room: 0 5 4 13  \n"
         + "room: 4 1 7   8\n"
         + "room : 7 4 13 8\n"
         + "room :   4 8 13 13\n"
         + "door: 3   2 4 2\n"
         + "door: 3 6   4 6\n"
         + "door  :   3   10   4   10  \n"
         + "door:  11   7   11  8\n"
         + "monster: 10   5 O \n"
         + "monster:   10 8   G\n"
         + "monster: 5   9 G\n"
         + "item: 5 4 A   50\n"
         + "item: 5 5 W   10\n"
         + "item: 5   9 A 50\n"
         + "item: 6 11   W 20\n"
         + "item: 11   8 W 30\n"
         + "item: 11 5  A 50\n"
        )
    };

    @Test
    public void testMapParsing()
    {
        DungeonMap dungeonMap = testMapParsingSub(DungeonParserTests.TEST_MAPS[0]);
        assertEquals(dungeonMap.getDescription(), "This is the real map!");
        testMapParsingSub(DungeonParserTests.TEST_MAPS[1]);
    }
        
    public DungeonMap testMapParsingSub(String map) {
        DungeonMap dungeonMap = DungeonParser.parseMap
            (new Scanner(map));

        assertEquals(dungeonMap.getScaleFactor(), 20);
        
        ArrayList<Rectangle> rooms = dungeonMap.getRooms();
        assertEquals(rooms.size(), 5);
        roomsEq(rooms.get(0), 0, 0, 4, 5);
        roomsEq(rooms.get(1), 0, 5, 4, 13);
        roomsEq(rooms.get(2), 4, 1, 7, 8);
        roomsEq(rooms.get(3), 7, 4, 13, 8);
        roomsEq(rooms.get(4), 4, 8, 13, 13);

        ArrayList<Point[]> doors = dungeonMap.getDoors();
        assertEquals(doors.size(), 4);
        doorsEq(doors.get(0), 3, 2, 4, 2);
        doorsEq(doors.get(1), 3, 6, 4, 6);
        doorsEq(doors.get(2), 3, 10, 4, 10);
        doorsEq(doors.get(3), 11, 7, 11, 8);

        ArrayList<Point> goblins = dungeonMap.getGoblins();
        assertEquals(goblins.size(), 2);
        monstersEq(goblins.get(0), 10, 8);
        monstersEq(goblins.get(1), 5, 9);

        ArrayList<Point> orcs = dungeonMap.getOrcs();
        assertEquals(orcs.size(), 1);
        monstersEq(orcs.get(0), 10, 5);

        ArrayList<DungeonMap.PositionAndStrength> armor
            = dungeonMap.getArmor();
        assertEquals(armor.size(), 3);
        itemsEq(armor.get(0), 5, 4, 50);
        itemsEq(armor.get(1), 5, 9, 50);
        itemsEq(armor.get(2), 11, 5, 50);

        ArrayList<DungeonMap.PositionAndStrength> weapons
            = dungeonMap.getWeapons();
        assertEquals(weapons.size(), 3);
        itemsEq(weapons.get(0), 5, 5, 10);
        itemsEq(weapons.get(1), 6, 11, 20);
        itemsEq(weapons.get(2), 11, 8, 30);

        Dungeon dungeon = dungeonMap.getDungeon();
        assertFalse(dungeon.isTooLarge());
        assertFalse(dungeon.isMalformed());

        return dungeonMap;
    }

    private void roomsEq(Rectangle rect, int x1, int y1, int x2, int y2)
    {
        assertEquals(rect.getUpperLeft().getX(), x1);
        assertEquals(rect.getUpperLeft().getY(), y1);
        assertEquals(rect.getLowerRight().getX(), x2);
        assertEquals(rect.getLowerRight().getY(), y2);
    }

    private void doorsEq(Point[] ps, int x1, int y1, int x2, int y2)
    {
        assertEquals(ps[0].getX(), x1);
        assertEquals(ps[0].getY(), y1);
        assertEquals(ps[1].getX(), x2);
        assertEquals(ps[1].getY(), y2);
    }

    private void monstersEq(Point p, int x, int y)
    {
        assertEquals(p.getX(), x);
        assertEquals(p.getY(), y);
    }

    private void itemsEq(DungeonMap.PositionAndStrength ps, int x, int y, int s)
    {
        assertEquals(ps.position.getX(), x);
        assertEquals(ps.position.getY(), y);
        assertEquals(ps.strength, s);
    }
}
